# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Regions helper functions.

Throughout Gammapy, we use `regions` to represent and work with regions.

https://astropy-regions.readthedocs.io

We might add in other conveniences and features here, e.g. sky coord contains
without a WCS (see "sky and pixel regions" in PIG 10), or some HEALPix integration.
"""

import operator
import numpy as np
from scipy.optimize import Bounds, minimize
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
from gammapy.utils.scripts import make_path
from regions import (
    CircleAnnulusSkyRegion,
    CircleSkyRegion,
    CompoundSkyRegion,
    EllipseSkyRegion,
    RectangleSkyRegion,
    Regions,
    PolygonSkyRegion,
    PolygonPixelRegion,
)

from regions.core.pixcoord import PixCoord
from regions.core.metadata import RegionMeta, RegionVisual
from regions._utils.wcs_helpers import pixel_scale_angle_at_skycoord

__all__ = [
    "compound_region_to_regions",
    "make_concentric_annulus_sky_regions",
    "make_orthogonal_rectangle_sky_regions",
    "regions_to_compound_region",
    "region_to_frame",
]


def compound_region_center(compound_region):
    """Compute center for a CompoundRegion.

    The center of the compound region is defined here as the geometric median
    of the individual centers of the regions. The geometric median is defined
    as the point the minimises the distance to all other points.

    Parameters
    ----------
    compound_region : `CompoundRegion`
        Compound region.

    Returns
    -------
    center : `~astropy.coordinates.SkyCoord`
        Geometric median of the positions of the individual regions.
    """
    regions = compound_region_to_regions(compound_region)

    if len(regions) == 1:
        return regions[0].center

    positions = SkyCoord([region.center.icrs for region in regions])

    def f(x, coords):
        """Function to minimize."""
        lon, lat = x
        center = SkyCoord(lon * u.deg, lat * u.deg)
        return np.sum(center.separation(coords).deg)

    ra, dec = positions.ra.wrap_at("180d").deg, positions.dec.deg

    ub = np.array([np.max(ra), np.max(dec)])
    lb = np.array([np.min(ra), np.min(dec)])

    if np.all(ub == lb):
        bounds = None
    else:
        bounds = Bounds(ub=ub, lb=lb)

    result = minimize(
        f,
        x0=[np.mean(ra), np.mean(dec)],
        args=(positions,),
        bounds=bounds,
        method="L-BFGS-B",
    )

    return SkyCoord(result.x[0], result.x[1], frame="icrs", unit="deg")


def compound_region_to_regions(region):
    """Create list of regions from compound regions.

    Parameters
    ----------
    region : `~regions.CompoundSkyRegion` or `~regions.SkyRegion`
        Compound sky region.

    Returns
    -------
    regions : `~regions.Regions`
        List of regions.
    """
    regions = Regions([])

    if isinstance(region, CompoundSkyRegion):
        if region.operator is operator.or_:
            regions_1 = compound_region_to_regions(region.region1)
            regions.extend(regions_1)

            regions_2 = compound_region_to_regions(region.region2)
            regions.extend(regions_2)
        else:
            raise ValueError("Only union operator supported")
    else:
        return Regions([region])

    return regions


def regions_to_compound_region(regions):
    """Create compound region from list of regions, by creating the union.

    Parameters
    ----------
    regions : `~regions.Regions`
        List of regions.

    Returns
    -------
    compound : `~regions.CompoundSkyRegion` or `~regions.CompoundPixelRegion`
        Compound sky region.
    """
    region_union = regions[0]

    for region in regions[1:]:
        region_union = region_union.union(region)

    return region_union


def get_centroid(vertices):
    """Compute centroid of a polygon. Implicitly assumes a flat
    cartesian projection, will probably break for very large polygons.

    Code comes from:
    https://stackoverflow.com/questions/75699024/finding-the-centroid-of-a-polygon-in-python

    Parameters
    ----------
    vertices : `~astropy.coordinates.SkyCoord`
        List of vertices.

    Returns
    -------
    centroid : `~astropy.coordinates.SkyCoord`
        Centroid of the polygon.
    """
    polygon = []
    for vertex in vertices:
        polygon.append((vertex.ra.degree, vertex.dec.degree))
    polygon = np.array(polygon)

    # Same polygon, but with vertices cycled around. Now the polygon
    # decomposes into triangles of the form origin-polygon[i]-polygon2[i]
    polygon2 = np.roll(polygon, -1, axis=0)

    # Compute signed area of each triangle
    signed_areas = 0.5 * np.cross(polygon, polygon2)

    # Compute centroid of each triangle
    centroids = (polygon + polygon2) / 3.0

    # Get average of those centroids, weighted by the signed areas.
    centroid = np.average(centroids, axis=0, weights=signed_areas)

    return SkyCoord(centroid[0] * u.deg, centroid[1] * u.deg, frame=vertices.frame)


class SphericalCircleSkyRegion(CircleSkyRegion):
    """Spherical circle sky region.

    TODO: is this separate class a good idea?

    - If yes, we could move it to the ``regions`` package?
    - If no, we should implement some other solution.
      Probably the alternative is to add extra methods to
      the ``CircleSkyRegion`` class and have that support
      both planar approximation and spherical case?
      Or we go with the approach to always make a
      TAN WCS and not have true cone select at all?
    """

    def contains(self, skycoord, wcs=None):
        """Defined by spherical distance."""
        separation = self.center.separation(skycoord)
        return separation < self.radius


class PolygonPointsSkyRegion(PolygonSkyRegion):
    """Polygon sky region defined by a list of points."""

    def __init__(self, vertices, meta=None, visual=None):
        """Create a polygon sky region.

        Parameters
        ----------
        vertices : `~astropy.coordinates.SkyCoord`
            List of vertices.
        meta : `~regions.RegionMeta`, optional
            Region meta data.
        visual : `~regions.RegionVisual`, optional
            Region visual meta data.
        """
        self.vertices = vertices
        self.meta = meta or RegionMeta()
        self.center = get_centroid(vertices)
        self.visual = visual or RegionVisual()

    def to_pixel(self, wcs):
        """Convert to pixel region."""
        x, y = wcs.world_to_pixel(self.vertices)
        center, _, _ = pixel_scale_angle_at_skycoord(self.center, wcs)

        vertices_pix = PixCoord(x, y)
        return PolygonPointsPixelRegion(
            vertices_pix,
            center=center,
            meta=self.meta.copy(),
            visual=self.visual.copy(),
        )


class PolygonPointsPixelRegion(PolygonPixelRegion):
    """Polygon pixel region defined by a list of points."""

    def __init__(self, vertices, center=None, meta=None, visual=None, origin=None):
        """Create a polygon pixel region.

        Parameters
        ----------
        vertices : `~regions.PixCoord`
            List of vertices.
        center : `~regions.PixCoord`, optional
            Center of the region.
        meta : `~regions.RegionMeta`, optional
            Region meta data.
        visual : `~regions.RegionVisual`, optional
            Region visual meta data.
        origin : `~regions.PixCoord`, optional
            Origin of the region. Default is `PixCoord(0, 0)`
        """

        if origin is None:
            origin = PixCoord(0, 0)

        self._vertices = vertices
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()
        self.origin = origin
        self.vertices = vertices + origin
        self.center = center

    def to_sky(self, wcs):
        """Convert to sky region.

        Parameters
        ----------
        wcs : `~astropy.wcs.WCS`
            WCS transformation object.

        """
        vertices_sky = wcs.pixel_to_world(self.vertices.x, self.vertices.y)

        return PolygonPointsSkyRegion(
            vertices=vertices_sky, meta=self.meta.copy(), visual=self.visual.copy()
        )

    def rotate(self, center, angle):
        """
        Rotate the region.

        Positive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `PolygonPixelRegion`
            The rotated region (which is an independent copy).
        """
        vertices = self.vertices.rotate(center, angle)
        center = self.center.rotate(center, angle)

        return self.copy(vertices=vertices, center=center)


def make_orthogonal_rectangle_sky_regions(start_pos, end_pos, wcs, height, nbin=1):
    """Utility returning an array of regions to make orthogonal projections.

    Parameters
    ----------
    start_pos : `~astropy.regions.SkyCoord`
        First sky coordinate defining the line to which the orthogonal boxes made.
    end_pos : `~astropy.regions.SkyCoord`
        Second sky coordinate defining the line to which the orthogonal boxes made.
    height : `~astropy.quantity.Quantity`
        Height of the rectangle region.
    wcs : `~astropy.wcs.WCS`
        WCS projection object.
    nbin : int, optional
        Number of boxes along the line. Default is 1.

    Returns
    -------
    regions : list of `~regions.RectangleSkyRegion`
        Regions in which the profiles are made.
    """
    pix_start = start_pos.to_pixel(wcs)
    pix_stop = end_pos.to_pixel(wcs)

    points = np.linspace(start=pix_start, stop=pix_stop, num=nbin + 1).T
    centers = 0.5 * (points[:, :-1] + points[:, 1:])
    coords = SkyCoord.from_pixel(centers[0], centers[1], wcs)

    width = start_pos.separation(end_pos).to("rad") / nbin
    angle = end_pos.position_angle(start_pos) - 90 * u.deg

    regions = []

    for center in coords:
        reg = RectangleSkyRegion(
            center=center, width=width, height=u.Quantity(height), angle=angle
        )
        regions.append(reg)

    return regions


def make_concentric_annulus_sky_regions(
    center, radius_max, radius_min=1e-5 * u.deg, nbin=11
):
    """Make a list of concentric annulus regions.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        Center coordinate.
    radius_max : `~astropy.units.Quantity`
        Maximum radius.
    radius_min : `~astropy.units.Quantity`, optional
        Minimum radius. Default is 1e-5 deg.
    nbin : int, optional
        Number of boxes along the line. Default is 11.

    Returns
    -------
    regions : list of `~regions.RectangleSkyRegion`
        Regions in which the profiles are made.
    """
    regions = []

    edges = np.linspace(radius_min, u.Quantity(radius_max), nbin)

    for r_in, r_out in zip(edges[:-1], edges[1:]):
        region = CircleAnnulusSkyRegion(
            center=center,
            inner_radius=r_in,
            outer_radius=r_out,
        )
        regions.append(region)

    return regions


def region_to_frame(region, frame):
    """Convert a region to a different frame.

    Parameters
    ----------
    region : `~regions.SkyRegion`
        Region to transform.
    frame : {"icrs", "galactic"}
        Frame to transform the region into.

    Returns
    -------
    region_new : `~regions.SkyRegion`
        Region in the given frame.
    """
    from gammapy.maps import WcsGeom

    wcs = WcsGeom.create(skydir=region.center, binsz=0.01, frame=frame).wcs
    region_new = region.to_pixel(wcs).to_sky(wcs)
    return region_new


def region_circle_to_ellipse(region):
    """Convert a CircleSkyRegion to an EllipseSkyRegion.

    Parameters
    ----------
    region : `~regions.CircleSkyRegion`
        Region to transform.

    Returns
    -------
    region_new : `~regions.EllipseSkyRegion`
        Elliptical region with same major and minor axis.
    """
    region_new = EllipseSkyRegion(
        center=region.center, width=region.radius, height=region.radius
    )
    return region_new


def extract_bright_star_regions(
    geom, max_star_mag=6, star_rad=0.3 * u.deg, star_table=None
):
    """
    Find bright star exclusion regions from a catalog `~astropy.table.Table`
    contained in a given `~gammapy.maps.Geom` and returns a list of `~regions.CircleSkyRegion`.

    Parameters
    ----------
    geom : `~gammapy.maps.WcsGeom`
        WcsGeom within which to search for bright stars
    max_star_mag : float, optional
        The maximum stellar magnitude (Johnson-Cousins B-band) to exclude. Default is 6th magnitude.
    star_rad : `~astropy.units.Quantity`, optional
        Radius to exclude around each star. Default is 0.3 degrees.
    star_table : str, optional
        Full name of the file containing a table containing the list of stars to search for exclusion regions.
        If None, the Hipparcos Catalog (Perryman et al., 1997) contained in ``$GAMMAPY_DATA//veritas/crab-point-like-ED`` is used.
        The table's column names should match those in ``$GAMMAPY_DATA//veritas/crab-point-like-ED``.
        Default is None.

    Returns
    -------
    regions : list of `~regions.CircleSkyRegion`
        Star exclusion regions.
    """
    regions = []

    if star_table is None:
        filename = make_path(
            "$GAMMAPY_DATA/veritas/crab-point-like-ED/Hipparcos_MAG8_1997.dat"
        )
        star_cat = Table.read(filename, format="ascii.commented_header")
    else:
        star_cat = Table.read(star_table)

    star_mask = geom.contains(
        SkyCoord(star_cat["_RAJ2000"] * u.deg, star_cat["_DEJ2000"] * u.deg)
    )

    for src in star_cat[
        (star_mask) & (star_cat["B-V"] + star_cat["Vmag"] < max_star_mag)
    ]:
        regions.append(
            CircleSkyRegion(
                center=SkyCoord(
                    src["_RAJ2000"], src["_DEJ2000"], unit="deg", frame="icrs"
                ),
                radius=star_rad,
            )
        )

    return regions



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import QTable, Table
from astropy.time import Time
from regions import CircleSkyRegion, RectangleSkyRegion
from gammapy.data import GTI, EventList, Observation, FixedPointingInfo
from gammapy.maps import MapAxis, WcsGeom
from gammapy.utils.testing import mpl_plot_check, requires_data


class TestEventListBasic:
    def setup_class(self):
        table = Table()
        table["RA"] = [0.0, 0.0, 0.0, 10.0] * u.deg
        table["DEC"] = [0.0, 0.9, 10.0, 10.0] * u.deg
        table["ENERGY"] = [1.0, 1.5, 1.5, 10.0] * u.TeV
        table["TIME"] = Time("2025-01-01") + [0.1, 0.5, 1.0, 1.5] * u.second

        self.events = EventList(table)

    def test_eventlist_printin(self):
        assert "Instrument" in str(self.events)


@pytest.fixture()
def simple_event_table():
    zeros = np.zeros(5)
    energy = zeros * u.TeV
    ra = zeros * u.deg
    dec = zeros * u.deg
    time = Time(zeros, format="mjd", scale="tt")
    return QTable({"ENERGY": energy, "RA": ra, "DEC": dec, "TIME": time})


def test_validation(simple_event_table):
    events = EventList(simple_event_table)
    assert len(events.table) == 5

    with pytest.raises(TypeError):
        EventList(simple_event_table["TIME"])

    bad = simple_event_table.copy()

    bad["ENERGY"] *= 1 * u.s
    with pytest.raises(u.UnitConversionError):
        EventList(bad)

    del bad["ENERGY"]
    with pytest.raises(KeyError):
        EventList(bad)


@requires_data()
class TestEventListBase:
    def setup_class(self):
        self.events = EventList.read(
            "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
        )

    def test_select_parameter(self):
        events = self.events.select_parameter("ENERGY", (0.8 * u.TeV, 5.0 * u.TeV))
        assert len(events.table) == 2716

        with pytest.warns(UserWarning):
            events = self.events.select_parameter(
                "ENERGY", (0.8, 5, 10) * u.TeV, is_range=True
            )
            assert len(events.table) == 2716

        events = self.events.select_parameter(
            "EVENT_ID", [1808181231761, 3594887627737, 3599182594792], is_range=False
        )
        assert len(events.table) == 3

        events = self.events.select_parameter("ENERGY", (0.8, np.inf) * u.TeV)
        assert len(events.table) == 3944

    def test_meta(self):
        assert self.events.meta.event_class == "std"
        assert self.events.meta.creation.creator == "SASH FITS::EventListWriter"
        assert self.events.meta.creation.date is None
        assert self.events.meta.creation.origin == "H.E.S.S. Collaboration"
        assert self.events.table["EVENT_ID"][0] == 1808181231761

    def test_write(self):
        # Without GTI and pointing
        obs = Observation(events=self.events)
        # Write function is through obs
        with pytest.raises(ValueError):
            obs.write("test.fits.gz", include_irfs=False, overwrite=True)

        pointing = FixedPointingInfo.from_fits_header(self.events.table.meta)
        obs = Observation(events=self.events, pointing=pointing)
        obs.write("test.fits.gz", include_irfs=False, overwrite=True)
        read_again = EventList.read("test.fits.gz")

        assert_allclose(self.events.table["RA"], read_again.table["RA"])
        assert_allclose(self.events.table["DEC"], read_again.table["DEC"])
        assert_allclose(self.events.table["ENERGY"], read_again.table["ENERGY"])
        assert_allclose(self.events.table["TIME"].mjd, read_again.table["TIME"].mjd)

        assert read_again.table.meta["EXTNAME"] == "EVENTS"
        assert read_again.table.meta["HDUCLASS"] == "GADF"
        assert read_again.table.meta["HDUCLAS1"] == "EVENTS"

        # With GTI
        gti = GTI.read(
            "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
        )

        obs = Observation(events=self.events, gti=gti, pointing=pointing)
        obs.write("test.fits", overwrite=True)
        read_again_ev = EventList.read("test.fits")
        read_again_gti = GTI.read("test.fits")

        assert_allclose(self.events.table["RA"], read_again_ev.table["RA"])
        assert_allclose(self.events.table["DEC"], read_again_ev.table["DEC"])
        assert_allclose(self.events.table["ENERGY"], read_again_ev.table["ENERGY"])
        assert_allclose(self.events.table["TIME"].mjd, read_again_ev.table["TIME"].mjd)
        assert gti.table.meta == read_again_gti.table.meta
        assert_allclose(gti.table["START"].mjd, read_again_gti.table["START"].mjd)
        assert_allclose(gti.table["STOP"].mjd, read_again_gti.table["STOP"].mjd)

        # test that it won't work if gti is not a GTI
        with pytest.raises(AttributeError):
            obs = Observation(events=self.events, gti=gti.table, pointing=pointing)
            obs.write("test.fits", overwrite=True)

    def test_eventlist_hdu_creation_metadata(self):
        hdu = self.events.to_table_hdu(format="gadf")
        assert "CREATOR" in hdu.header
        assert "CREATED" in hdu.header
        assert hdu.header["CREATOR"] == "SASH FITS::EventListWriter"

    def test_eventlist_read_kwargs(self):
        with pytest.raises(ValueError):
            EventList.read(
                "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz",
                format="unknown-format",
            )


@requires_data()
class TestEventListHESS:
    def setup_class(self):
        self.events = EventList.read(
            "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
        )

    def test_basics(self):
        assert "EventList" in str(self.events)

        assert self.events.is_pointed_observation

        assert len(self.events.table) == 11243
        assert self.events.time[0].iso == "2004-03-26 02:57:47.004"
        assert self.events.radec[0].to_string() == "229.239 -58.3417"
        assert self.events.galactic[0].to_string(precision=2) == "321.07 -0.69"
        assert self.events.altaz[0].to_string() == "193.338 53.258"
        assert_allclose(self.events.offset[0].value, 0.54000974, rtol=1e-5)

        energy = self.events.energy[0]
        assert energy.unit == "TeV"
        assert_allclose(energy.value, 0.55890286)

        lon, lat, height = self.events.observatory_earth_location.to_geodetic()
        assert lon.unit == "deg"
        assert_allclose(lon.value, 16.5002222222222)
        assert lat.unit == "deg"
        assert_allclose(lat.value, -23.2717777777778)
        assert height.unit == "m"
        assert_allclose(height.value, 1835)

    def test_observation_time_duration(self):
        dt = self.events.observation_time_duration
        assert dt.unit == "s"
        assert_allclose(dt.value, 1682)

    def test_observation_live_time_duration(self):
        dt = self.events.observation_live_time_duration
        assert dt.unit == "s"
        assert_allclose(dt.value, 1521.026855)

    def test_observation_dead_time_fraction(self):
        deadc = self.events.observation_dead_time_fraction
        assert_allclose(deadc, 0.095703, rtol=1e-3)

    def test_altaz(self):
        altaz = self.events.altaz
        assert_allclose(altaz[0].az.deg, 193.337965, atol=1e-3)
        assert_allclose(altaz[0].alt.deg, 53.258024, atol=1e-3)

    def test_median_position(self):
        coord = self.events.galactic_median
        assert_allclose(coord.l.deg, 320.539346, atol=1e-3)
        assert_allclose(coord.b.deg, -0.882515, atol=1e-3)

    def test_median_offset(self):
        offset_max = self.events.offset_from_median.max()
        assert_allclose(offset_max.to_value("deg"), 36.346379, atol=1e-3)

    def test_from_stack(self):
        event_lists = [self.events] * 2
        stacked_list = EventList.from_stack(event_lists)
        assert len(stacked_list.table) == 11243 * 2

    def test_stack(self):
        events, other = self.events.copy(), self.events.copy()
        events.stack(other)
        assert len(events.table) == 11243 * 2

    def test_offset_selection(self):
        offset_range = u.Quantity([0.5, 1.0] * u.deg)
        new_list = self.events.select_offset(offset_range)
        assert len(new_list.table) == 1820

    def test_plot_time(self):
        with mpl_plot_check():
            self.events.plot_time()

    def test_plot_energy(self):
        with mpl_plot_check():
            self.events.plot_energy()

    def test_plot_offset2_distribution(self):
        with mpl_plot_check():
            self.events.plot_offset2_distribution()

    def test_plot_energy_offset(self):
        with mpl_plot_check():
            self.events.plot_energy_offset()

    def test_plot_image(self):
        with mpl_plot_check():
            self.events.plot_image()

    def test_peek(self):
        with mpl_plot_check():
            self.events.peek()


@requires_data()
class TestEventListFermi:
    def setup_class(self):
        self.events = EventList.read(
            "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc-events.fits.gz"
        )

    def test_basics(self):
        assert "EventList" in str(self.events)
        assert len(self.events.table) == 32843
        assert not self.events.is_pointed_observation

    def test_peek(self):
        with mpl_plot_check():
            self.events.peek(allsky=True)


@pytest.mark.xfail
@requires_data()
class TestEventListChecker:
    def setup_class(self):
        self.event_list = EventList.read(
            "$GAMMAPY_DATA/cta-1dc/data/baseline/gps/gps_baseline_111140.fits"
        )

    def test_check_all(self):
        records = list(self.event_list.check())
        assert len(records) == 3


class TestEventSelection:
    def setup_class(self):
        table = Table()
        table["RA"] = [0.0, 0.0, 0.0, 10.0] * u.deg
        table["DEC"] = [0.0, 0.9, 10.0, 10.0] * u.deg
        table["ENERGY"] = [1.0, 1.5, 1.5, 10.0] * u.TeV
        table["OFFSET"] = [0.1, 0.5, 1.0, 1.5] * u.deg
        table["TIME"] = Time("2025-01-01") + [0.1, 0.5, 1.0, 1.5] * u.s

        self.events = EventList(table)

        center1 = SkyCoord(0.0, 0.0, frame="icrs", unit="deg")
        on_region1 = CircleSkyRegion(center1, radius=1.0 * u.deg)
        center2 = SkyCoord(0.0, 10.0, frame="icrs", unit="deg")
        on_region2 = RectangleSkyRegion(center2, width=0.5 * u.deg, height=0.3 * u.deg)
        self.on_regions = [on_region1, on_region2]

    def test_region_select(self):
        geom = WcsGeom.create(skydir=(0, 0), binsz=0.2, width=4.0 * u.deg, proj="TAN")
        new_list = self.events.select_region(self.on_regions[0], geom.wcs)
        assert len(new_list.table) == 2

        union_region = self.on_regions[0].union(self.on_regions[1])
        new_list = self.events.select_region(union_region, geom.wcs)
        assert len(new_list.table) == 3

        region_string = "fk5;box(0,10, 0.25, 0.15)"
        new_list = self.events.select_region(region_string, geom.wcs)
        assert len(new_list.table) == 1

    def test_map_select(self):
        axis = MapAxis.from_edges((0.5, 2.0), unit="TeV", name="ENERGY")
        geom = WcsGeom.create(
            skydir=(0, 0), binsz=0.2, width=4.0 * u.deg, proj="TAN", axes=[axis]
        )

        mask = geom.region_mask(regions=[self.on_regions[0]])
        new_list = self.events.select_mask(mask)
        assert len(new_list.table) == 2

    def test_select_energy(self):
        energy_range = u.Quantity([1, 10], "TeV")
        new_list = self.events.select_energy(energy_range)
        assert len(new_list.table) == 3



r"""
.. _piecewise-norm-spectral:

Piecewise  norm spectral model
==============================

This model parametrises a piecewise spectral correction
with a free norm parameter at each fixed energy node.
"""

# %%
# Example plot
# ------------
# Here is an example plot of the model:

from astropy import units as u
import matplotlib.pyplot as plt
from gammapy.modeling.models import (
    Models,
    PiecewiseNormSpectralModel,
    PowerLawSpectralModel,
    SkyModel,
)

energy_bounds = [0.1, 100] * u.TeV
model = PiecewiseNormSpectralModel(
    energy=[0.1, 1, 3, 10, 30, 100] * u.TeV,
    norms=[1, 3, 8, 10, 8, 2],
)
model.plot(energy_bounds, yunits=u.Unit(""))
plt.grid(which="both")


# %%
# YAML representation
# -------------------
# Here is an example YAML file using the model:

model = model * PowerLawSpectralModel()
model = SkyModel(spectral_model=model, name="piecewise-norm-model")
models = Models([model])

print(models.to_yaml())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from astropy.units import Quantity
from gammapy.modeling.models import PowerLawSpectralModel
from gammapy.utils.integrate import trapz_loglog
from gammapy.utils.testing import assert_quantity_allclose


def test_trapz_loglog():
    energy = Quantity([1, 10], "TeV")
    pwl = PowerLawSpectralModel(index=2.3)

    ref = pwl.integral(energy_min=energy[0], energy_max=energy[1])

    val = trapz_loglog(pwl(energy), energy)
    assert_quantity_allclose(val, ref)



from astropy.coordinates import Angle, SkyCoord
from regions import CircleSkyRegion
import matplotlib.pyplot as plt
from gammapy.makers import ReflectedRegionsFinder
from gammapy.maps import RegionGeom, WcsNDMap

# Exclude a rectangular region
exclusion_mask = WcsNDMap.create(npix=(801, 701), binsz=0.01, skydir=(83.6, 23.0))

coords = exclusion_mask.geom.get_coord().skycoord
data = (Angle("23 deg") < coords.dec) & (coords.dec < Angle("24 deg"))
exclusion_mask.data = ~data

pos = SkyCoord(83.633, 22.014, unit="deg")
radius = Angle(0.3, "deg")
on_region = CircleSkyRegion(pos, radius)
center = SkyCoord(83.633, 24, unit="deg")

# One can impose a minimal distance between ON region and first reflected regions
finder = ReflectedRegionsFinder(
    min_distance_input="0.2 rad",
)
regions, _ = finder.run(
    region=on_region,
    center=center,
    exclusion_mask=exclusion_mask,
)

fig, axes = plt.subplots(
    ncols=3,
    subplot_kw={"projection": exclusion_mask.geom.wcs},
    figsize=(12, 3),
)


def plot_regions(ax, regions, on_region, exclusion_mask):
    """Little helper function to plot off regions"""
    exclusion_mask.plot_mask(ax=ax, colors="gray")
    on_region.to_pixel(ax.wcs).plot(ax=ax, edgecolor="tab:orange")
    geom = RegionGeom.from_regions(regions)
    geom.plot_region(ax=ax, color="tab:blue")


ax = axes[0]
ax.set_title("Min. distance first region")
plot_regions(ax=ax, regions=regions, on_region=on_region, exclusion_mask=exclusion_mask)


# One can impose a minimal distance between two adjacent regions
finder = ReflectedRegionsFinder(
    min_distance="0.1 rad",
)
regions, _ = finder.run(
    region=on_region,
    center=center,
    exclusion_mask=exclusion_mask,
)

ax = axes[1]
ax.set_title("Min. distance all regions")
plot_regions(ax=ax, regions=regions, on_region=on_region, exclusion_mask=exclusion_mask)


# One can impose a maximal number of regions to be extracted
finder = ReflectedRegionsFinder(
    max_region_number=5,
    min_distance="0.1 rad",
)
regions, _ = finder.run(
    region=on_region,
    center=center,
    exclusion_mask=exclusion_mask,
)

ax = axes[2]
ax.set_title("Max. number of regions")
plot_regions(ax=ax, regions=regions, on_region=on_region, exclusion_mask=exclusion_mask)
plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import abc
import numpy as np
from pathlib import Path
from astropy import units as u
from astropy.io import fits
from astropy.table import Table
from gammapy.data import GTI
from gammapy.irf import EDispKernel, EDispKernelMap, PSFMap
from gammapy.maps import RegionNDMap, Map
from gammapy.modeling.models import create_fermi_isotropic_diffuse_model, Models
from gammapy.utils.scripts import read_yaml, make_path
from gammapy.utils.metadata import CreatorMetaData
from .spectrum import SpectrumDatasetOnOff
from .utils import create_map_dataset_from_dl4

__all__ = [
    "DatasetReader",
    "DatasetWriter",
    "OGIPDatasetReader",
    "OGIPDatasetWriter",
    "FermipyDatasetsReader",
]


class DatasetReader(abc.ABC):
    """Dataset reader base class."""

    @property
    @abc.abstractmethod
    def tag(self):
        pass

    @abc.abstractmethod
    def read(self):
        pass


class DatasetWriter(abc.ABC):
    """Dataset writer base class."""

    @property
    @abc.abstractmethod
    def tag(self):
        pass

    @abc.abstractmethod
    def write(self, dataset):
        pass


class OGIPDatasetWriter(DatasetWriter):
    """Write OGIP files.

    If you want to use the written files with Sherpa, you have to use the
    ``ogip-sherpa`` format. Then all files will be written in units of 'keV' and
    'cm2'.

    The naming scheme is fixed as following:

    * PHA file is named filename.fits
    * BKG file is named filename_bkg.fits
    * ARF file is named filename_arf.fits
    * RMF file is named filename_rmf.fits

    Parameters
    ----------
    filename : `~pathlib.Path` or str
        Filename.
    format : {"ogip", "ogip-sherpa"}
        Which format to use. Default is 'ogip'.
    overwrite : bool, optional
        Overwrite existing files. Default is False.
    checksum : bool
        When True adds both DATASUM and CHECKSUM cards to the headers written to the files.
        Default is False.
    creation : `~gammapy.utils.metadata.CreatorMetaData`, optional.
        The creation metadata to write to disk. If None, use default creator metadata object.
    """

    tag = ["ogip", "ogip-sherpa"]

    def __init__(
        self, filename, format="ogip", overwrite=False, checksum=False, creation=None
    ):
        filename = make_path(filename)
        filename.parent.mkdir(exist_ok=True, parents=True)

        self.filename = filename
        self.format = format
        self.overwrite = overwrite
        self.checksum = checksum
        self.creation = creation or CreatorMetaData()

    @staticmethod
    def get_filenames(filename):
        """Get filenames.

        Parameters
        ----------
        filename : `~pathlib.Path`
            Filename.

        Returns
        -------
        filenames : dict
            Dictionary of filenames.
        """
        suffix = "".join(filename.suffixes)
        name = filename.name.replace(suffix, "")
        name = f"{name}{{}}{suffix}"
        return {
            "respfile": name.format("_rmf"),
            "backfile": name.format("_bkg"),
            "ancrfile": name.format("_arf"),
        }

    def get_ogip_meta(self, dataset, is_bkg=False):
        """Meta info for the OGIP data format."""
        try:
            livetime = dataset.exposure.meta["livetime"]
        except KeyError:
            raise ValueError(
                "Storing in ogip format require the livetime "
                "to be defined in the exposure meta data"
            )

        hdu_class = "BKG" if is_bkg else "TOTAL"

        meta = {
            "HDUCLAS2": hdu_class,
            "HDUCLAS3": "COUNT",
            "HDUCLAS4": "TYPE:1",
            "EXPOSURE": livetime.to_value("s"),
            "OBS_ID": dataset.name,
        }

        filenames = OGIPDatasetWriter.get_filenames(self.filename)
        meta["ANCRFILE"] = filenames["ancrfile"]

        if dataset.counts_off:
            meta["BACKFILE"] = filenames["backfile"]

        if dataset.edisp:
            meta["RESPFILE"] = filenames["respfile"]

        return meta

    def write(self, dataset):
        """Write dataset to file.

        Parameters
        ----------
        dataset : `SpectrumDatasetOnOff`
            Dataset to write.
        """
        filenames = self.get_filenames(self.filename)

        self.creation.update_time()

        self.write_pha(dataset, filename=self.filename)

        path = self.filename.parent
        self.write_arf(dataset, filename=path / filenames["ancrfile"])

        if dataset.counts_off:
            self.write_bkg(dataset, filename=path / filenames["backfile"])

        if dataset.edisp:
            self.write_rmf(dataset, filename=path / filenames["respfile"])

    def write_rmf(self, dataset, filename):
        """Write energy dispersion.

        Parameters
        ----------
        dataset : `SpectrumDatasetOnOff`
            Dataset to write.
        filename : str or `~pathlib.Path`
            Filename to use.
        """
        kernel = dataset.edisp.get_edisp_kernel()
        kernel.write(
            filename=filename,
            format=self.format,
            checksum=self.checksum,
            overwrite=self.overwrite,
        )

    def write_arf(self, dataset, filename):
        """Write effective area.

        Parameters
        ----------
        dataset : `SpectrumDatasetOnOff`
            Dataset to write.
        filename : str or `~pathlib.Path`
            Filename to use.

        """
        aeff = dataset.exposure / dataset.exposure.meta["livetime"]

        filename = make_path(filename)
        hdulist = aeff.to_hdulist(format=self.format.replace("ogip", "ogip-arf"))

        for hdu in hdulist:
            hdu.header.update(self.creation.to_header())

        hdulist.writeto(filename, overwrite=self.overwrite, checksum=self.checksum)

    def to_counts_hdulist(self, dataset, is_bkg=False):
        """Convert counts region map to hdulist.

        Parameters
        ----------
        dataset : `SpectrumDatasetOnOff`
            Dataset to write.
        is_bkg : bool
            Whether to use counts off. Default is False.
        """
        counts = dataset.counts_off if is_bkg else dataset.counts
        acceptance = dataset.acceptance_off if is_bkg else dataset.acceptance

        hdulist = counts.to_hdulist(format=self.format)

        table = Table.read(hdulist["SPECTRUM"])
        meta = self.get_ogip_meta(dataset, is_bkg=is_bkg)

        if dataset.mask_safe is not None:
            mask_array = dataset.mask_safe.data[:, 0, 0]
        else:
            mask_array = np.ones(acceptance.data.size)

        table["QUALITY"] = np.logical_not(mask_array)
        del table.meta["QUALITY"]

        table["BACKSCAL"] = acceptance.data[:, 0, 0]
        del table.meta["BACKSCAL"]

        # adapt meta data
        table.meta.update(meta)
        hdulist["SPECTRUM"] = fits.BinTableHDU(table)

        for hdu in hdulist:
            hdu.header.update(self.creation.to_header())
        return hdulist

    def write_pha(self, dataset, filename):
        """Write counts file.

        Parameters
        ----------
        dataset : `SpectrumDatasetOnOff`
            Dataset to write.
        filename : str or `~pathlib.Path`
            Filename to use.

        """
        hdulist = self.to_counts_hdulist(dataset)

        if dataset.gti:
            hdu = dataset.gti.to_table_hdu()
            hdulist.append(hdu)

        hdulist.writeto(filename, overwrite=self.overwrite, checksum=self.checksum)

    def write_bkg(self, dataset, filename):
        """Write off counts file.

        Parameters
        ----------
        dataset : `SpectrumDatasetOnOff`
            Dataset to write.
        filename : str or `~pathlib.Path`
            Filename to use.
        """
        hdulist = self.to_counts_hdulist(dataset, is_bkg=True)
        hdulist.writeto(filename, overwrite=self.overwrite, checksum=self.checksum)


class OGIPDatasetReader(DatasetReader):
    """Read `~gammapy.datasets.SpectrumDatasetOnOff` from OGIP files.

    BKG file, ARF, and RMF must be set in the PHA header and be present in
    the same folder.

    The naming scheme is fixed to the following scheme:

    * PHA file is named ``pha_obs{name}.fits``
    * BKG file is named ``bkg_obs{name}.fits``
    * ARF file is named ``arf_obs{name}.fits``
    * RMF file is named ``rmf_obs{name}.fits``
      with ``{name}`` the dataset name.

    Parameters
    ----------
    filename : str or `~pathlib.Path`
        OGIP PHA file to read.
    checksum : bool
        If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.
    """

    tag = "ogip"

    def __init__(self, filename, checksum=False, name=None):
        self.filename = make_path(filename)
        self.checksum = checksum
        self.name = name

    def get_valid_path(self, filename):
        """Get absolute or relative path.

        The relative path is with respect to the name of the reference file.

        Parameters
        ----------
        filename : str or `~pathlib.Path`
            Filename.

        Returns
        -------
        filename : `~pathlib.Path`
            Valid path.
        """
        filename = make_path(filename)

        if not filename.exists():
            return self.filename.parent / filename
        else:
            return filename

    def get_filenames(self, pha_meta):
        """Get filenames.

        Parameters
        ----------
        pha_meta : dict
            Metadata from the PHA file.

        Returns
        -------
        filenames : dict
            Dictionary with filenames of "arffile", "rmffile" (optional)
            and "bkgfile" (optional).
        """
        filenames = {"arffile": self.get_valid_path(pha_meta["ANCRFILE"])}

        if "BACKFILE" in pha_meta:
            filenames["bkgfile"] = self.get_valid_path(pha_meta["BACKFILE"])

        if "RESPFILE" in pha_meta:
            filenames["rmffile"] = self.get_valid_path(pha_meta["RESPFILE"])

        return filenames

    @staticmethod
    def read_pha(filename, checksum=False):
        """Read PHA file.

        Parameters
        ----------
        filename : str or `~pathlib.Path`
            PHA file name.
        checksum : bool
            If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.

        Returns
        -------
        data : dict
            Dictionary with counts, acceptance and mask_safe.
        """
        data = {}

        with fits.open(filename, memmap=False, checksum=checksum) as hdulist:
            data["counts"] = RegionNDMap.from_hdulist(hdulist, format="ogip")
            data["acceptance"] = RegionNDMap.from_hdulist(
                hdulist, format="ogip", ogip_column="BACKSCAL"
            )

            if "GTI" in hdulist:
                data["gti"] = GTI.from_table_hdu(hdulist["GTI"])

            data["mask_safe"] = RegionNDMap.from_hdulist(
                hdulist, format="ogip", ogip_column="QUALITY"
            )

        return data

    @staticmethod
    def read_bkg(filename, checksum=False):
        """Read PHA background file.

        Parameters
        ----------
        filename : str or `~pathlib.Path`
            PHA file name.
        checksum : bool
            If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.

        Returns
        -------
        data : dict
            Dictionary with counts_off and acceptance_off.
        """
        with fits.open(filename, memmap=False, checksum=checksum) as hdulist:
            counts_off = RegionNDMap.from_hdulist(hdulist, format="ogip")
            acceptance_off = RegionNDMap.from_hdulist(
                hdulist, ogip_column="BACKSCAL", format="ogip"
            )
        return {"counts_off": counts_off, "acceptance_off": acceptance_off}

    @staticmethod
    def read_rmf(filename, exposure, checksum=False):
        """Read RMF file.

        Parameters
        ----------
        filename : str or `~pathlib.Path`
            PHA file name.
        exposure : `RegionNDMap`
            Exposure map.
        checksum : bool
            If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.

        Returns
        -------
        data : `EDispKernelMap`
            Dictionary with edisp.
        """
        kernel = EDispKernel.read(filename, checksum=checksum)
        edisp = EDispKernelMap.from_edisp_kernel(kernel, geom=exposure.geom)

        # TODO: resolve this separate handling of exposure for edisp
        edisp.exposure_map.data = exposure.data[:, :, np.newaxis, :]
        return edisp

    @staticmethod
    def read_arf(filename, livetime, checksum=False):
        """Read ARF file.

        Parameters
        ----------
        filename : str or `~pathlib.Path`
            PHA file name.
        livetime : `Quantity`
            Livetime.
        checksum : bool
            If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.

        Returns
        -------
        data : `RegionNDMap`
            Exposure map.
        """
        aeff = RegionNDMap.read(filename, format="ogip-arf", checksum=checksum)
        exposure = aeff * livetime
        exposure.meta["livetime"] = livetime
        return exposure

    def read(self):
        """Read dataset.

        Returns
        -------
        dataset : SpectrumDatasetOnOff
            Spectrum dataset.
        """
        kwargs = self.read_pha(self.filename, checksum=self.checksum)
        pha_meta = kwargs["counts"].meta

        if self.name is not None:
            name = self.name
        else:
            name = str(pha_meta["OBS_ID"])
        livetime = pha_meta["EXPOSURE"] * u.s

        filenames = self.get_filenames(pha_meta=pha_meta)
        exposure = self.read_arf(
            filenames["arffile"], livetime=livetime, checksum=self.checksum
        )

        if "bkgfile" in filenames:
            bkg = self.read_bkg(filenames["bkgfile"], checksum=self.checksum)
            kwargs.update(bkg)

        if "rmffile" in filenames:
            kwargs["edisp"] = self.read_rmf(
                filenames["rmffile"], exposure=exposure, checksum=self.checksum
            )

        return SpectrumDatasetOnOff(name=name, exposure=exposure, **kwargs)


class FermipyDatasetsReader(DatasetReader):
    """Create datasets from Fermi-LAT files.

    Parameters
    ----------
    filename : str
        Path to Fermipy configuration file (tested only for v1.3.1).
    edisp_bins : int
        Number of margin bins to slice in energy. Default is 0.
        For now only maps created with edisp_bins=0 in fermipy configuration are supported,
        in that case the emin/emax in the fermipy configuration will correspond to the true energy range for gammapy,
        and  a value edisp_bins>0 should be set here in order to apply the energy dispersion correctly.
        With a binning of 8 to 10 bins per decade, it is recommended to use edisp_bins ≥ 2
        (See https://fermi.gsfc.nasa.gov/ssc/data/analysis/documentation/Pass8_edisp_usage.html)

    """

    tag = "fermipy"

    def __init__(self, filename, edisp_bins=0):
        self.filename = make_path(filename)
        self.edisp_bins = edisp_bins

    @staticmethod
    def create_dataset(
        counts_file,
        exposure_file,
        psf_file,
        edisp_file,
        isotropic_file=None,
        edisp_bins=0,
        name=None,
    ):
        """Create a map dataset from Fermi-LAT files.

        Parameters
        ----------
        counts_file : str
            Counts file path.
        exposure_file : str
            Exposure file path.
        psf_file : str
            Point spread function file path.
        edisp_file : str
            Energy dispersion file path.
        isotropic_file : str, optional
            Isotropic file path. Default is None
        edisp_bins : int
            Number of margin bins to slice in energy. Default is 0.
            For now only maps created with edisp_bins=0 in fermipy configuration are supported,
            in that case the emin/emax in the fermipy configuration will correspond to the true energy range for gammapy,
            and  a value edisp_bins>0 should be set here in order to apply the energy dispersion correctly.
            With a binning of 8 to 10 bins per decade, it is recommended to use edisp_bins ≥ 2
            (See https://fermi.gsfc.nasa.gov/ssc/data/analysis/documentation/Pass8_edisp_usage.html)
        name : str, optional
            Dataset name. The default is None, and the name is randomly generated.

        Returns
        -------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset.

        """
        from gammapy.datasets import MapDataset

        counts = Map.read(counts_file)
        exposure = Map.read(exposure_file)
        psf = PSFMap.read(psf_file, format="gtpsf")
        edisp = EDispKernelMap.read(edisp_file, format="gtdrm")

        # check that fermipy edisp_bins are matching between edisp and exposure
        # as we will interp to edisp axis the exposure axis must be larger or equal
        edisp_axes = edisp.edisp_map.geom.axes
        if len(edisp_axes["energy_true"].center) > len(
            exposure.geom.axes["energy_true"].center
        ):
            raise ValueError(
                "Energy true axes of exposure and DRM do not match. Check fermipy configuration."
            )
            edisp_axes = edisp.edisp_map.geom.axes

        psf_r68s = psf.containment_radius(
            0.68,
            edisp_axes["energy_true"].center,
            position=counts.geom.center_skydir,
        )
        # check that pdf is well defined (fails if edisp_bins>0 in fermipy)
        if np.any(psf_r68s.value) == 0.0:
            raise ValueError(
                "PSF is not defined for all true energies. Check fermipy configuration."
            )

        # change counts energy axis unit keV->MeV
        energy_axis = counts.geom.axes["energy"]._init_copy(
            nodes=edisp_axes["energy"].edges
        )
        geom = counts.geom.to_image().to_cube([energy_axis])
        counts = Map.from_geom(geom, data=counts.data)

        # standardize dataset interpolating to same geom and axes
        dataset = MapDataset(
            counts=counts,
            exposure=exposure,
            psf=psf,
            edisp=edisp,
            name=name,
        )
        dataset = create_map_dataset_from_dl4(
            dataset, geom=counts.geom, name=dataset.name
        )

        if edisp_bins > 0:  # slice edisp_bins
            dataset = dataset.slice_by_idx(
                dict(energy=slice(edisp_bins, -edisp_bins)), name=dataset.name
            )

        if isotropic_file:
            model = create_fermi_isotropic_diffuse_model(
                isotropic_file, datasets_names=[dataset.name]
            )
            dataset.models = Models([model])
        return dataset

    def read(self):
        """Create Fermi-LAT map datasets from Fermipy configuration file.

        Returns
        -------
        dataset : `~gammapy.datasets.Datasets`
            Map datasets.

        """
        from gammapy.datasets import Datasets

        filename = self.filename.resolve()
        data = read_yaml(filename)

        if "components" in data:
            components = data["components"]
        else:
            components = [data]

        datasets = Datasets()
        for file_id, component in enumerate(components):
            if "fileio" in component and "outdir" in component["fileio"]:
                path = Path(component["fileio"]["outdir"])
            elif "fileio" in data and "outdir" in data["fileio"]:
                path = Path(data["fileio"]["outdir"])
            else:
                path = Path(filename.parent)
            if not path.is_absolute():
                path = Path(filename.parent) / path

            if "model" in component and "isodiff" in component["model"]:
                isotropic_file = Path(component["model"]["isodiff"])
                name = isotropic_file.stem[4:]
            elif "model" in data and "isodiff" in data["model"]:
                isotropic_file = Path(data["model"]["isodiff"])
                name = isotropic_file.stem[4:]
            else:
                isotropic_file = None
                name = None

            datasets.append(
                self.create_dataset(
                    counts_file=path / f"ccube_0{str(file_id)}.fits",
                    exposure_file=path / f"bexpmap_0{str(file_id)}.fits",
                    psf_file=path / f"psf_0{str(file_id)}.fits",
                    edisp_file=path / f"drm_0{str(file_id)}.fits",
                    isotropic_file=isotropic_file,
                    edisp_bins=self.edisp_bins,
                    name=name,
                )
            )
        return datasets



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import astropy.units as u
from astropy.table import Table
from gammapy.utils.cluster import standard_scaler

__all__ = ["get_irfs_features"]


def get_irfs_features(
    observations,
    energy_true,
    position=None,
    fixed_offset=None,
    names=None,
    containment_fraction=0.68,
    apply_standard_scaler=False,
):
    """Get features from IRFs properties at a given position. Used for observations clustering.

    Parameters
    ----------
    observations : `~gammapy.data.Observations`
        Container holding a list of `~gammapy.data.Observation`.
    energy_true : `~astropy.units.Quantity`
        Energy true at which to compute the containment radius.
    position : `~astropy.coordinates.SkyCoord`, optional
        Sky position. Default is None.
    fixed_offset : `~astropy.coordinates.Angle`, optional
        Offset calculated from the pointing position. Default is None.
        If neither the `position` nor the `fixed_offset` is specified,
        it uses the position of the center of the map by default.
    names : {"edisp-bias", "edisp-res", "psf-radius"}
        IRFs properties to be considered.
        Default is None. If None, all the features are computed.
    containment_fraction : float, optional
        Containment_fraction to compute the `psf-radius`.
        Default is 68%.
    apply_standard_scaler : bool, optional
        Compute standardize features by removing the mean and scaling to unit variance.
        Default is False.

    Returns
    -------
    features : `~astropy.table.Table`
        Features table.

    Examples
    --------
    Compute the IRF features for "edisp-bias", "edisp-res" and "psf-radius" at 1 TeV::

    >>> from gammapy.data.utils import get_irfs_features
    >>> from gammapy.data import DataStore
    >>> from gammapy.utils.cluster import standard_scaler
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u

    >>> selection = dict(
    ...     type="sky_circle",
    ...     frame="icrs",
    ...     lon="329.716 deg",
    ...     lat="-30.225 deg",
    ...     radius="2 deg",
    ... )

    >>> data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")
    >>> obs_table = data_store.obs_table.select_observations(selection)
    >>> obs = data_store.get_observations(obs_table["OBS_ID"][:3])

    >>> position = SkyCoord(329.716 * u.deg, -30.225 * u.deg, frame="icrs")
    >>> names = ["edisp-bias", "edisp-res", "psf-radius"]
    >>> features_irfs = get_irfs_features(
    ...     obs,
    ...     energy_true="1 TeV",
    ...     position=position,
    ...     names=names,
    ... )

    >>> print(features_irfs)
         edisp-bias     obs_id      edisp-res           psf-radius
                                                       deg
        ------------------- ------ ------------------- -------------------
        0.11587179071752986  33787   0.368346217294295 0.14149953611195087
        0.04897634344908595  33788 0.33983991887701287 0.11553325504064559
          0.033176650892097  33789 0.32377509405904137 0.10262943822890519

    """
    from gammapy.irf import EDispKernelMap, PSFMap

    if names is None:
        names = ["edisp-bias", "edisp-res", "psf-radius"]

    if position and fixed_offset:
        raise ValueError(
            "`position` and `fixed_offset` arguments are mutually exclusive"
        )

    rows = []

    for obs in observations:
        psf_kwargs = dict(fraction=containment_fraction, energy_true=energy_true)
        if isinstance(obs.psf, PSFMap) and isinstance(obs.edisp, EDispKernelMap):
            if position is None:
                position = obs.psf.psf_map.geom.center_skydir
            edisp_kernel = obs.edisp.get_edisp_kernel(position=position)
            psf_kwargs["position"] = position
        else:
            if fixed_offset is None:
                if position is None:
                    offset = 0 * u.deg
                else:
                    offset_max = np.minimum(
                        obs.psf.axes["offset"].center[-1],
                        obs.edisp.axes["offset"].center[-1],
                    )
                    offset = np.minimum(
                        position.separation(obs.get_pointing_icrs(obs.tmid)), offset_max
                    )
            else:
                offset = fixed_offset
            edisp_kernel = obs.edisp.to_edisp_kernel(offset)
            psf_kwargs["offset"] = offset

        data = {}
        for name in names:
            if name == "edisp-bias":
                data[name] = edisp_kernel.get_bias(energy_true)[0]
            if name == "edisp-res":
                data[name] = edisp_kernel.get_resolution(energy_true)[0]
            if name == "psf-radius":
                data[name] = obs.psf.containment_radius(**psf_kwargs).to("deg")
            data["obs_id"] = obs.obs_id

        rows.append(data)

    features = Table(rows)

    if apply_standard_scaler:
        features = standard_scaler(features)

    features.meta = dict(
        energy_true=energy_true,
        fixed_offset=fixed_offset,
        containment_fraction=containment_fraction,
        apply_standard_scaler=apply_standard_scaler,
    )

    if position:
        features.meta["lon"] = position.galactic.l
        features.meta["lat"] = position.galactic.b
        features.meta["frame"] = "galactic"

    return features



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import pytest
from numpy.testing import assert_allclose
from gammapy.modeling.fit import Fit
from gammapy.modeling.models import Models
from gammapy.modeling.selection import NestedModelSelection, select_nested_models
from gammapy.utils.testing import requires_data


@pytest.fixture(scope="session")
def fermi_datasets():
    from gammapy.datasets import Datasets

    filename = "$GAMMAPY_DATA/fermi-3fhl-crab/Fermi-LAT-3FHL_datasets.yaml"
    filename_models = "$GAMMAPY_DATA/fermi-3fhl-crab/Fermi-LAT-3FHL_models.yaml"
    fermi_datasets = Datasets.read(filename=filename, filename_models=filename_models)
    return fermi_datasets


@requires_data()
def test_test_statistic_detection(fermi_datasets):
    model = fermi_datasets.models["Crab Nebula"]

    results = select_nested_models(
        fermi_datasets, [model.spectral_model.amplitude], [0]
    )
    assert_allclose(results["ts"], 20905.667798, rtol=1e-5)
    assert fermi_datasets.models.parameters["amplitude"].error != 0.0

    ts_eval = NestedModelSelection([model.spectral_model.amplitude], [0])
    ts_known_bkg = ts_eval.ts_known_bkg(fermi_datasets)
    ts_asimov = ts_eval.ts_asimov(fermi_datasets)
    ts = ts_eval.ts(fermi_datasets)
    assert_allclose(ts_known_bkg, 28086.269146, rtol=1e-2)
    assert_allclose(ts_asimov, 27495.908291, rtol=1e-2)
    assert_allclose(ts, 20905.667798, rtol=1e-5)

    # bad model
    bias_factor = 1.2
    model.spectral_model.amplitude.value *= bias_factor
    ts_eval = NestedModelSelection([model.spectral_model.amplitude], [0])
    ts_known_bkg = ts_eval.ts_known_bkg(fermi_datasets)
    ts_asimov = ts_eval.ts_asimov(fermi_datasets)
    ts = ts_eval.ts(fermi_datasets)
    assert_allclose(ts_known_bkg, 28002.894113, rtol=1e-2)
    assert_allclose(ts_asimov, 34130.645399, rtol=1e-2)  # ts_known_bkg<<ts_asimov
    assert_allclose(ts, 20905.667798, rtol=1e-5)
    model.spectral_model.amplitude.value /= bias_factor


@requires_data()
def test_test_statistic_detection_other_frozen(fermi_datasets):
    with fermi_datasets.models.restore_status():
        fermi_datasets.models.freeze()
        model = fermi_datasets.models["Crab Nebula"]
        results = select_nested_models(
            fermi_datasets, [model.spectral_model.amplitude], [0]
        )
        results["fit_results_null"].nfev == 0
        assert fermi_datasets.models.parameters["amplitude"].error != 0.0

        model.spectral_model.amplitude.value = 0
        assert_allclose(
            results["fit_results_null"].parameters.value,
            fermi_datasets.models.parameters.value,
        )


@requires_data()
def test_test_statistic_link(fermi_datasets):
    # TODO: better test with simulated data ?

    models = Models.read("$GAMMAPY_DATA/fermi-3fhl-crab/Fermi-LAT-3FHL_models.yaml")

    model = models["Crab Nebula"]
    model2 = model.copy(name="other")
    model2.spectral_model.alpha.value = 2.4

    fermi_datasets.models = models + [model2]

    fit = Fit()
    minuit_opts = {"tol": 10, "strategy": 0}
    fit.backend = "minuit"
    fit.optimize_opts = minuit_opts

    ts_eval = NestedModelSelection(
        [model.spectral_model.alpha], [model2.spectral_model.alpha], fit=fit
    )
    results = ts_eval.run(fermi_datasets)

    assert results["ts"] > ts_eval.ts_threshold
    assert model2.spectral_model.alpha.value != model.spectral_model.alpha.value
    assert model2.spectral_model.alpha.error != model.spectral_model.alpha.error
    assert model2.spectral_model.alpha.error != 0

    ts_eval = NestedModelSelection(
        [model.spectral_model.alpha],
        [model2.spectral_model.alpha],
        fit=fit,
        n_sigma=np.inf,
    )
    results = ts_eval.run(fermi_datasets)

    assert results["ts"] < ts_eval.ts_threshold
    assert_allclose(model2.spectral_model.alpha.value, model.spectral_model.alpha.value)
    assert_allclose(model2.spectral_model.alpha.error, model.spectral_model.alpha.error)
    assert model2.spectral_model.alpha.error != 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from pathlib import Path
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from regions import (
    CircleAnnulusSkyRegion,
    CircleSkyRegion,
    EllipseSkyRegion,
    PointSkyRegion,
    RectangleSkyRegion,
)
from gammapy.maps import Map, MapAxis, MapCoord, RegionGeom, WcsGeom, WcsNDMap
from gammapy.modeling.models import (
    SPATIAL_MODEL_REGISTRY,
    ConstantSpatialModel,
    DiskSpatialModel,
    FoVBackgroundModel,
    GaussianSpatialModel,
    GeneralizedGaussianSpatialModel,
    PiecewiseNormSpatialModel,
    PointSpatialModel,
    PowerLawSpectralModel,
    Shell2SpatialModel,
    ShellSpatialModel,
    SkyModel,
    TemplateNDSpatialModel,
    TemplateSpatialModel,
)
from gammapy.utils.testing import mpl_plot_check, requires_data, requires_dependency


def test_sky_point_source():
    geom = WcsGeom.create(skydir=(2.4, 2.3), npix=(10, 10), binsz=0.3)
    model = PointSpatialModel(lon_0="2.5 deg", lat_0="2.5 deg", frame="icrs")

    assert model.is_energy_dependent is False

    assert model.evaluation_radius.unit == "deg"
    assert_allclose(model.evaluation_radius.value, 0)

    assert model.frame == "icrs"

    assert_allclose(model.position.ra.deg, 2.5)
    assert_allclose(model.position.dec.deg, 2.5)

    val = model.evaluate_geom(geom)
    assert val.unit == "sr-1"
    assert_allclose(np.sum(val * geom.solid_angle()), 1)

    assert isinstance(model.to_region(), PointSkyRegion)


def test_sky_gaussian():
    # Test symmetric model
    sigma = 1 * u.deg
    model = GaussianSpatialModel(lon_0="5 deg", lat_0="15 deg", sigma=sigma)
    assert model.parameters["sigma"].min == 0
    val_0 = model(5 * u.deg, 15 * u.deg)
    val_sigma = model(5 * u.deg, 16 * u.deg)
    assert val_0.unit == "sr-1"
    ratio = val_0 / val_sigma
    assert_allclose(ratio, np.exp(0.5))
    radius = model.evaluation_radius
    assert radius.unit == "deg"
    assert_allclose(radius.value, 5 * sigma.value)
    assert_allclose(model.evaluation_bin_size_min, (1.0 / 3.0) * u.deg)

    # test the normalization for an elongated Gaussian near the Galactic Plane
    m_geom_1 = WcsGeom.create(
        binsz=0.05, width=(20, 20), skydir=(2, 2), frame="galactic", proj="AIT"
    )
    coords = m_geom_1.get_coord()
    solid_angle = m_geom_1.solid_angle()
    lon = coords.lon
    lat = coords.lat
    sigma = 3 * u.deg
    model_1 = GaussianSpatialModel(
        lon_0=2 * u.deg, lat_0=2 * u.deg, sigma=sigma, e=0.8, phi=30 * u.deg
    )
    vals_1 = model_1(lon, lat)
    assert vals_1.unit == "sr-1"
    assert_allclose(np.sum(vals_1 * solid_angle), 1, rtol=1.0e-3)

    radius = model_1.evaluation_radius
    assert radius.unit == "deg"
    assert_allclose(radius.value, 5 * sigma.value)

    # check the ratio between the value at the peak and on the 1-sigma isocontour
    sigma = 4 * u.deg
    semi_minor = 2 * u.deg
    e = np.sqrt(1 - (semi_minor / sigma) ** 2)
    model_2 = GaussianSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, sigma=sigma, e=e, phi=0 * u.deg
    )
    val_0 = model_2(0 * u.deg, 0 * u.deg)
    val_major = model_2(0 * u.deg, 4 * u.deg)
    val_minor = model_2(2 * u.deg, 0 * u.deg)
    assert val_0.unit == "sr-1"
    ratio_major = val_0 / val_major
    ratio_minor = val_0 / val_minor

    assert_allclose(ratio_major, np.exp(0.5))
    assert_allclose(ratio_minor, np.exp(0.5))

    # check the rotation
    model_3 = GaussianSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, sigma=sigma, e=e, phi=90 * u.deg
    )
    val_minor_rotated = model_3(0 * u.deg, 2 * u.deg)
    ratio_minor_rotated = val_0 / val_minor_rotated
    assert_allclose(ratio_minor_rotated, np.exp(0.5))

    assert isinstance(model.to_region(), EllipseSkyRegion)


@pytest.mark.parametrize("eta", np.arange(0.1, 1.01, 0.3))
@pytest.mark.parametrize("r_0", np.arange(0.01, 1.01, 0.3))
@pytest.mark.parametrize("e", np.arange(0.0, 0.801, 0.4))
def test_generalized_gaussian(eta, r_0, e):
    # check normalization is robust for a large set of values
    model = GeneralizedGaussianSpatialModel(
        eta=eta, r_0=r_0 * u.deg, e=e, frame="galactic"
    )

    width = np.maximum(2 * model.evaluation_radius.to_value("deg"), 0.5)
    geom = WcsGeom.create(
        skydir=(0, 0),
        binsz=0.02,
        width=width,
        frame="galactic",
    )

    integral = model.integrate_geom(geom)
    assert integral.unit.is_equivalent("")
    assert_allclose(integral.data.sum(), 1.0, atol=5e-3)


def test_generalized_gaussian_io():
    model = GeneralizedGaussianSpatialModel(e=0.5)

    reg = model.to_region()
    assert isinstance(reg, EllipseSkyRegion)
    assert_allclose(reg.width.value, 1.73205, rtol=1e-5)

    new_model = GeneralizedGaussianSpatialModel.from_dict(model.to_dict())
    assert isinstance(new_model, GeneralizedGaussianSpatialModel)


def test_sky_disk():
    # Test the disk case (e=0)
    r_0 = 2 * u.deg
    model = DiskSpatialModel(lon_0="1 deg", lat_0="45 deg", r_0=r_0)
    lon = [1, 5, 359] * u.deg
    lat = 46 * u.deg
    val = model(lon, lat)
    assert val.unit == "sr-1"
    desired = [261.263956, 0, 261.263956]
    assert_allclose(val.value, desired)
    radius = model.evaluation_radius
    assert radius.unit == "deg"
    assert_allclose(radius.to_value("deg"), 2.222)
    assert_allclose(model.evaluation_bin_size_min, 0.198 * u.deg)

    assert model.phi.min == -180
    assert model.phi.max == 180

    model.phi.quantity = 210 * u.deg
    val1 = model(lon, lat)
    model.phi.value = 30
    val2 = model(lon, lat)
    assert_allclose(val1, val2)

    # test the normalization for an elongated ellipse near the Galactic Plane
    m_geom_1 = WcsGeom.create(
        binsz=0.015, width=(20, 20), skydir=(2, 2), frame="galactic", proj="AIT"
    )
    coords = m_geom_1.get_coord()
    solid_angle = m_geom_1.solid_angle()
    lon = coords.lon
    lat = coords.lat
    r_0 = 10 * u.deg
    model_1 = DiskSpatialModel(
        lon_0=2 * u.deg, lat_0=2 * u.deg, r_0=r_0, e=0.4, phi=30 * u.deg
    )
    vals_1 = model_1(lon, lat)
    assert vals_1.unit == "sr-1"
    assert_allclose(np.sum(vals_1 * solid_angle), 1, rtol=1.0e-3)

    radius = model_1.evaluation_radius
    assert radius.unit == "deg"
    assert_allclose(radius.to_value("deg"), 11.11)
    # test rotation
    r_0 = 2 * u.deg
    semi_minor = 1 * u.deg
    eccentricity = np.sqrt(1 - (semi_minor / r_0) ** 2)
    model_rot_test = DiskSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, r_0=r_0, e=eccentricity, phi=90 * u.deg
    )
    assert_allclose(model_rot_test(0 * u.deg, 1.5 * u.deg).value, 0)

    # test the normalization for a disk (ellipse with e=0) at the Galactic Pole
    m_geom_2 = WcsGeom.create(
        binsz=0.1, width=(6, 6), skydir=(0, 90), frame="galactic", proj="AIT"
    )
    coords = m_geom_2.get_coord()
    lon = coords.lon
    lat = coords.lat

    r_0 = 5 * u.deg
    disk = DiskSpatialModel(lon_0=0 * u.deg, lat_0=90 * u.deg, r_0=r_0)
    vals_disk = disk(lon, lat)

    solid_angle = 2 * np.pi * (1 - np.cos(5 * u.deg))
    assert_allclose(np.max(vals_disk).value * solid_angle, 1)

    assert isinstance(model.to_region(), EllipseSkyRegion)


def test_sky_disk_edge():
    r_0 = 2 * u.deg
    model = DiskSpatialModel(
        lon_0="0 deg",
        lat_0="0 deg",
        r_0=r_0,
        e=0.5,
        phi="0 deg",
    )
    value_center = model(0 * u.deg, 0 * u.deg)
    value_edge = model(0 * u.deg, r_0)
    assert_allclose((value_edge / value_center).to_value(""), 0.5)

    edge = model.edge_width.value * r_0
    value_edge_pwidth = model(0 * u.deg, r_0 + edge / 2)
    assert_allclose((value_edge_pwidth / value_center).to_value(""), 0.05)

    value_edge_nwidth = model(0 * u.deg, r_0 - edge / 2)
    assert_allclose((value_edge_nwidth / value_center).to_value(""), 0.95)


def test_disk_from_region():
    region = EllipseSkyRegion(
        center=SkyCoord(20, 17, unit="deg"),
        height=0.3 * u.deg,
        width=1.0 * u.deg,
        angle=30 * u.deg,
    )
    disk = DiskSpatialModel.from_region(region, frame="galactic")
    assert_allclose(disk.parameters["lon_0"].value, 132.666, rtol=1e-2)
    assert_allclose(disk.parameters["lat_0"].value, -45.33118067, rtol=1e-2)
    assert_allclose(disk.parameters["r_0"].quantity, 0.5 * u.deg, rtol=1e-2)
    assert_allclose(disk.parameters["e"].value, 0.9539, rtol=1e-2)
    assert_allclose(disk.parameters["phi"].quantity, 110.946048 * u.deg)

    reg1 = disk.to_region()
    assert_allclose(reg1.height, region.width, rtol=1e-2)

    center = SkyCoord(20, 17, unit="deg", frame="galactic")
    region = EllipseSkyRegion(
        center=center,
        height=1 * u.deg,
        width=0.3 * u.deg,
        angle=30 * u.deg,
    )
    disk = DiskSpatialModel.from_region(region, frame="icrs")
    reg1 = disk.to_region()
    assert_allclose(reg1.angle, -30.323 * u.deg, rtol=1e-2)
    assert_allclose(reg1.height, region.height, rtol=1e-3)

    region = CircleSkyRegion(center=region.center, radius=1.0 * u.deg)
    disk = DiskSpatialModel.from_region(region)
    assert_allclose(disk.parameters["e"].value, 0.0, rtol=1e-2)
    assert_allclose(disk.parameters["lon_0"].value, 20, rtol=1e-2)
    assert disk.frame == "galactic"

    geom = WcsGeom.create(skydir=center, npix=(10, 10), binsz=0.3)
    res = disk.evaluate_geom(geom)
    assert_allclose(np.sum(res.value), 50157.904662)

    region = PointSkyRegion(center=region.center)
    with pytest.raises(ValueError):
        DiskSpatialModel.from_region(region)


def test_from_position():
    center = SkyCoord(20, 17, unit="deg")
    spatial_model = GaussianSpatialModel.from_position(
        position=center, sigma=0.5 * u.deg
    )
    geom = WcsGeom.create(skydir=center, npix=(10, 10), binsz=0.3)
    res = spatial_model.evaluate_geom(geom)
    assert_allclose(np.sum(res.value), 36307.440813)
    model = SkyModel(
        spectral_model=PowerLawSpectralModel(), spatial_model=spatial_model
    )
    assert_allclose(model.position.ra.value, center.ra.value, rtol=1e-3)


def test_sky_shell():
    width = 2 * u.deg
    rad = 2 * u.deg
    model = ShellSpatialModel(lon_0="1 deg", lat_0="45 deg", radius=rad, width=width)
    lon = [1, 2, 4] * u.deg
    lat = 45 * u.deg
    val = model(lon, lat)
    assert val.unit == "deg-2"
    desired = [55.979449, 57.831651, 94.919895]
    assert_allclose(val.to_value("sr-1"), desired)
    radius = model.evaluation_radius
    assert radius.unit == "deg"
    assert_allclose(radius.value, rad.value + width.value)
    assert isinstance(model.to_region(), CircleAnnulusSkyRegion)
    assert_allclose(model.evaluation_bin_size_min, 2 * u.deg)


def test_sky_shell2():
    width = 2 * u.deg
    rad = 2 * u.deg
    model = Shell2SpatialModel(lon_0="1 deg", lat_0="45 deg", r_0=rad + width, eta=0.5)
    lon = [1, 2, 4] * u.deg
    lat = 45 * u.deg
    val = model(lon, lat)
    assert val.unit == "deg-2"
    desired = [55.979449, 57.831651, 94.919895]
    assert_allclose(val.to_value("sr-1"), desired)
    radius = model.evaluation_radius
    assert radius.unit == "deg"
    assert_allclose(radius.value, rad.value + width.value)
    assert_allclose(model.r_in.value, rad.value)
    assert isinstance(model.to_region(), CircleAnnulusSkyRegion)
    assert_allclose(model.evaluation_bin_size_min, 2 * u.deg)


def test_sky_diffuse_constant():
    model = ConstantSpatialModel(value="42 sr-1")
    lon = [1, 2] * u.deg
    lat = 45 * u.deg
    val = model(lon, lat)
    assert val.unit == "sr-1"
    assert_allclose(val.value, 42)
    radius = model.evaluation_radius
    assert radius is None
    assert isinstance(model.to_region(), RectangleSkyRegion)


@requires_data()
@requires_dependency("ipywidgets")
def test_sky_diffuse_map(caplog):
    filename = "$GAMMAPY_DATA/catalogs/fermi/Extended_archive_v18/Templates/RXJ1713_2016_250GeV.fits"  # noqa: E501
    model = TemplateSpatialModel.read(filename, normalize=False)
    lon = [258.5, 0] * u.deg
    lat = -39.8 * u.deg
    val = model(lon, lat)

    assert "WARNING" in [_.levelname for _ in caplog.records]
    assert "Missing spatial template unit, assuming sr^-1" in [
        _.message for _ in caplog.records
    ]

    assert val.unit == "sr-1"
    desired = [3265.6559, 0]
    assert_allclose(val.value, desired)

    res = model.evaluate_geom(model.map.geom)
    assert_allclose(np.sum(res.value), 32826159.74707)
    radius = model.evaluation_radius

    assert radius.unit == "deg"
    assert_allclose(radius.value, 0.64, rtol=1.0e-2)
    assert model.frame == "fk5"
    assert isinstance(model.to_region(), RectangleSkyRegion)

    with pytest.raises(TypeError):
        model.plot_interactive()

    with pytest.raises(TypeError):
        model.plot_grid()

    # change central position
    model.lon_0.value = 12.0
    model.lat_0.value = 6
    val = model([11.8, 12.8] * u.deg, 6.1 * u.deg)
    assert_allclose(val.value, [2850.8103, 89.629447], rtol=1e-3)

    # test to and from dict
    dict1 = model.to_dict()
    model2 = TemplateSpatialModel.from_dict(dict1)
    assert_allclose(model2.lon_0.quantity, 12.0 * u.deg, rtol=1e-3)

    # test dict without parameters
    dict1["spatial"]["parameters"] = []
    model3 = TemplateSpatialModel.from_dict(dict1)
    assert_allclose(model3.lon_0.quantity, 258.388 * u.deg, rtol=1e-3)

    # test dict without parameters
    dict1["spatial"].pop("parameters")
    model3 = TemplateSpatialModel.from_dict(dict1)
    assert_allclose(model3.lon_0.quantity, 258.388 * u.deg, rtol=1e-3)


@requires_data()
@requires_dependency("ipywidgets")
def test_sky_diffuse_map_3d():
    filename = "$GAMMAPY_DATA/fermi-3fhl-gc/gll_iem_v06_gc.fits.gz"
    model = TemplateSpatialModel.read(filename, normalize=False)
    lon = [258.5, 0] * u.deg
    lat = -39.8 * u.deg
    energy = 1 * u.GeV
    val = model(lon, lat, energy)

    with pytest.raises(ValueError):
        model(lon, lat)
    assert model.map.unit == "cm-2 s-1 MeV-1 sr-1"

    val = model(lon, lat, energy)
    assert val.unit == "cm-2 s-1 MeV-1 sr-1"

    res = model.evaluate_geom(model.map.geom)
    assert_allclose(np.sum(res.value), 0.11803847221522712)

    with pytest.raises(TypeError):
        model.plot()

    with mpl_plot_check():
        model.plot_grid()

    with mpl_plot_check():
        model.plot_interactive()


def test_sky_diffuse_map_normalize():
    # define model map with a constant value of 1
    model_map = Map.create(map_type="wcs", width=(10, 5), binsz=0.5, unit="sr-1")
    model_map.data += 1.0
    model = TemplateSpatialModel(model_map)

    # define data map with a different spatial binning
    data_map = Map.create(map_type="wcs", width=(10, 5), binsz=1)
    coords = data_map.geom.get_coord()
    solid_angle = data_map.geom.solid_angle()
    vals = model(coords.lon, coords.lat) * solid_angle

    assert vals.unit == ""
    integral = vals.sum()
    assert_allclose(integral.value, 1, rtol=1e-4)


def test_sky_diffuse_map_copy():
    # define model map with a constant value of 1
    model_map = Map.create(map_type="wcs", width=(1, 1), binsz=0.5, unit="sr-1")
    model_map.data += 1.0

    model = TemplateSpatialModel(model_map, normalize=False)
    assert np.all(model.map.data == model_map.data)
    model.map.data += 1
    # Check that the original map is unchanged
    assert np.all(model_map.data == np.ones_like(model_map.data))

    model = TemplateSpatialModel(model_map, normalize=False, copy_data=False)
    assert np.all(model.map.data == model_map.data)
    model.map.data += 1
    # Check that the original map has also been changed
    assert np.all(model.map.data == model_map.data)

    model_copy = model.copy(copy_data=False)
    model_copy.map.data += 1
    # Check that the original map has also been changed
    assert np.all(model.map.data == model_copy.map.data)


def test_sky_diffuse_map_empty(caplog):
    # define model map with 0 values
    model_map = Map.create(map_type="wcs", width=(1, 1), binsz=0.5, unit="sr-1")

    with caplog.at_level(logging.WARNING):
        model = TemplateSpatialModel(model_map, normalize=True)
        assert "Map values are all zeros. Check and fix this!" in [
            _.message for _ in caplog.records
        ]
        assert np.all(np.isfinite(model.map.data))

    axes = [MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=2)]
    model_map = Map.create(
        map_type="wcs", width=(1, 1), binsz=0.5, unit="sr-1", axes=axes
    )
    model_map.data[0, :, :] = 1
    with caplog.at_level(logging.WARNING):
        model = TemplateSpatialModel(model_map, normalize=True)
        assert (
            "Map values are all zeros in at least one energy bin. Check and fix this!"
            in [_.message for _ in caplog.records]
        )
        assert np.all(np.isfinite(model.map.data))

    # model with nan
    model_map.data[0, 0, 0] = np.nan
    with caplog.at_level(logging.WARNING):
        TemplateSpatialModel(model_map, normalize=True)
        assert "Map has NaN values. Check and fix this!" in [
            _.message for _ in caplog.records
        ]


@pytest.mark.parametrize("model_cls", SPATIAL_MODEL_REGISTRY)
def test_model_from_dict(tmpdir, model_cls):
    if model_cls in [TemplateSpatialModel, TemplateNDSpatialModel]:
        default_map = Map.create(map_type="wcs", width=(1, 1), binsz=0.5, unit="sr-1")
        filename = str(tmpdir / "template.fits")
        model = model_cls(default_map, filename=filename)
        model.write()

    elif model_cls is PiecewiseNormSpatialModel:
        geom = WcsGeom.create(skydir=(0, 0), npix=(2, 2), binsz=0.3, frame="galactic")
        default_coords = MapCoord.create(geom.footprint)
        default_coords["lon"] *= u.deg
        default_coords["lat"] *= u.deg
        model = model_cls(default_coords, frame="galactic")
    else:
        model = model_cls()

    data = model.to_dict()
    model_from_dict = model_cls.from_dict(data)
    assert model_from_dict.tag == model_from_dict.tag

    bkg_model = FoVBackgroundModel(spatial_model=model, dataset_name="test")
    bkg_model_dict = bkg_model.to_dict()
    bkg_model_from_dict = FoVBackgroundModel.from_dict(bkg_model_dict)
    assert bkg_model_from_dict.spatial_model is not None


def test_evaluate_on_fk5_map():
    # Check if spatial model can be evaluated on a map with FK5 frame
    # Regression test for GH-2402
    header = {}
    header["CDELT1"] = 1.0
    header["CDELT2"] = 1.0
    header["CTYPE1"] = "RA---TAN"
    header["CTYPE2"] = "DEC--TAN"
    header["RADESYS"] = "FK5"
    header["CRVAL1"] = 0
    header["CRVAL2"] = 0
    header["CRPIX1"] = 5
    header["CRPIX2"] = 5

    wcs = WCS(header)
    geom = WcsGeom(wcs, npix=(10, 10))
    model = GaussianSpatialModel(lon_0="0 deg", lat_0="0 deg", sigma="1 deg")
    data = model.evaluate_geom(geom)
    assert data.sum() > 0


def test_evaluate_fk5_model():
    geom = WcsGeom.create(width=(5, 5), binsz=0.1, frame="icrs")
    model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma="0.1 deg", frame="fk5"
    )
    data = model.evaluate_geom(geom)
    assert data.sum() > 0


def test_spatial_model_plot():
    model = PointSpatialModel()
    model.covariance = np.diag([0.01, 0.01])

    with mpl_plot_check():
        ax = model.plot()

    with mpl_plot_check():
        model.plot_error(ax=ax, which="all")


models_test = [
    (GaussianSpatialModel, "sigma"),
    (GeneralizedGaussianSpatialModel, "r_0"),
    (DiskSpatialModel, "r_0"),
]


@pytest.mark.parametrize(("model_class", "extension_param"), models_test)
def test_spatial_model_plot_error(model_class, extension_param):
    model = model_class(lon_0="0 deg", lat_0="0 deg", frame="galactic")
    model.lat_0.error = 0.04
    model.lon_0.error = 0.02
    model.parameters[extension_param].value = 0.2
    model.parameters[extension_param].error = 0.04
    model.e.error = 0.002

    empty_map = Map.create(
        skydir=model.position, frame=model.frame, width=1, binsz=0.02
    )
    with mpl_plot_check():
        ax = empty_map.plot()
        model.plot_error(ax=ax, which="all")
        model.plot_error(ax=ax, which="position")
        model.plot_error(ax=ax, which="extension")


def test_integrate_region_geom():
    center = SkyCoord("0d", "0d", frame="icrs")
    model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma=0.1 * u.deg, frame="icrs"
    )

    radius_large = 1 * u.deg
    circle_large = CircleSkyRegion(center, radius_large)
    radius_small = 0.1 * u.deg
    circle_small = CircleSkyRegion(center, radius_small)

    geom_large, geom_small = (
        RegionGeom(region=circle_large),
        RegionGeom(region=circle_small, binsz_wcs="0.01 deg"),
    )

    integral_large, integral_small = (
        model.integrate_geom(geom_large).data,
        model.integrate_geom(geom_small).data,
    )

    assert_allclose(integral_large[0], 1, rtol=0.001)
    assert_allclose(integral_small[0], 0.3953, rtol=0.001)


# TODO: solve issue with small radii (e.g. 1e-5) and improve tolerance
@pytest.mark.parametrize("width", np.geomspace(1e-4, 1e-1, 10) * u.deg)
@pytest.mark.parametrize(
    "model",
    [
        (GaussianSpatialModel, "sigma", 6e-3),
        (DiskSpatialModel, "r_0", 0.4),
        (GeneralizedGaussianSpatialModel, "r_0", 3e-4),
    ],
)
def test_integrate_wcs_geom(width, model):
    model_cls, param_name, tolerance = model
    param_dict = {param_name: width}
    spatial_model = model_cls(
        lon_0="0.234 deg", lat_0="-0.172 deg", frame="icrs", **param_dict
    )
    geom = WcsGeom.create(skydir=(0, 0), npix=100, binsz=0.02)

    integrated = spatial_model.integrate_geom(geom)

    assert_allclose(integrated.data.sum(), 1, atol=tolerance)


def test_integrate_geom_no_overlap():
    center = SkyCoord("0d", "0d", frame="icrs")
    model = GaussianSpatialModel(
        lon_0="10.234 deg", lat_0="-0.172 deg", sigma=1e-2 * u.deg, frame="icrs"
    )
    geom = WcsGeom.create(skydir=center, npix=100, binsz=0.02)

    # This should not fail but return map filled with 0
    integrated = model.integrate_geom(geom)

    assert_allclose(integrated.data, 0.0)


def test_integrate_geom_parameter_issue():
    center = SkyCoord("0d", "0d", frame="icrs")
    model = GaussianSpatialModel(
        lon_0="0.234 deg", lat_0="-0.172 deg", sigma=np.nan * u.deg, frame="icrs"
    )
    geom = WcsGeom.create(skydir=center, npix=100, binsz=0.02)

    # This should not fail but return map filled with nan
    integrated = model.integrate_geom(geom)

    assert_allclose(integrated.data, np.nan)


def test_integrate_geom_energy_axis():
    center = SkyCoord("0d", "0d", frame="icrs")
    model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma=0.1 * u.deg, frame="icrs"
    )

    radius = 1 * u.deg
    square = RectangleSkyRegion(center, radius, radius)

    axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=10)
    geom = RegionGeom(region=square, axes=[axis])

    integral = model.integrate_geom(geom).data

    assert_allclose(integral, 1, rtol=0.0001)


def test_templatemap_clip():
    model_map = Map.create(map_type="wcs", width=(2, 2), binsz=0.5, unit="sr-1")
    model_map.data += 1.0
    model = TemplateSpatialModel(model_map)
    model.map.data = model.map.data * -1

    lon = np.array([0, 0.2, 0.3]) * u.deg
    lat = np.array([0, 0.2, 0.3]) * u.deg

    val = model.evaluate(lon, lat)
    assert_allclose(val, 0, rtol=0.0001)


def test_piecewise_spatial_model_gc():
    geom = WcsGeom.create(skydir=(0, 0), npix=(2, 2), binsz=0.3, frame="galactic")
    coords = MapCoord.create(geom.footprint)
    coords["lon"] *= u.deg
    coords["lat"] *= u.deg

    model = PiecewiseNormSpatialModel(coords, frame="galactic")

    assert_allclose(model(*geom.to_image().center_coord), 1.0)

    norms = np.arange(coords.shape[0])

    model = PiecewiseNormSpatialModel(coords, norms, frame="galactic")

    assert not model.is_energy_dependent

    expected = np.array([[0, 3], [1, 2]])
    assert_allclose(model(*geom.to_image().get_coord()), expected, atol=1e-5)

    assert_allclose(model.evaluate_geom(geom.to_image()), expected, atol=1e-5)

    assert_allclose(model.evaluate_geom(geom), expected, atol=1e-5)

    model_dict = model.to_dict()
    new_model = PiecewiseNormSpatialModel.from_dict(model_dict)
    assert model_dict == new_model.to_dict()

    assert_allclose(new_model.evaluate_geom(geom.to_image()), expected, atol=1e-5)

    assert_allclose(
        model.evaluate(-0.1 * u.deg, 2.3 * u.deg),
        model.evaluate(359.9 * u.deg, 2.3 * u.deg),
    )


def test_piecewise_spatial_model():
    for lon in range(-360, 360):
        geom = WcsGeom.create(
            skydir=(lon, 2.3), npix=(2, 2), binsz=0.3, frame="galactic"
        )

        coords = MapCoord.create(geom.footprint)
        coords["lon"] *= u.deg
        coords["lat"] *= u.deg

        model = PiecewiseNormSpatialModel(coords, frame="galactic")

        assert_allclose(model(*geom.to_image().center_coord), 1.0)

        norms = np.arange(coords.shape[0])

        model = PiecewiseNormSpatialModel(coords, norms, frame="galactic")

        expected = np.array([[0, 3], [1, 2]])
        assert_allclose(model(*geom.to_image().get_coord()), expected, atol=1e-5)

        assert_allclose(model.evaluate_geom(geom.to_image()), expected, atol=1e-5)

        assert_allclose(model.evaluate_geom(geom), expected, atol=1e-5)

        model_dict = model.to_dict()
        new_model = PiecewiseNormSpatialModel.from_dict(model_dict)
        assert model_dict == new_model.to_dict()

        assert_allclose(new_model.evaluate_geom(geom.to_image()), expected, atol=1e-5)


def test_piecewise_spatial_model_3d():
    axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=3)
    geom = WcsGeom.create(
        skydir=(2.4, 2.3), npix=(2, 2), binsz=0.3, frame="galactic", axes=[axis]
    )
    coords = geom.get_coord().flat

    with pytest.raises(ValueError):
        PiecewiseNormSpatialModel(coords, frame="galactic")


@requires_data()
def test_template_ND(tmpdir, caplog):
    filename = "$GAMMAPY_DATA/catalogs/fermi/Extended_archive_v18/Templates/RXJ1713_2016_250GeV.fits"  # noqa: E501
    map_ = Map.read(filename)
    map_.data[map_.data < 0] = 0
    geom2d = map_.geom
    norm = MapAxis.from_nodes(range(0, 11, 2), interp="lin", name="norm", unit="")
    cste = MapAxis.from_bounds(-1, 1, 3, interp="lin", name="cste", unit="")
    geom = geom2d.to_cube([norm, cste])

    nd_map = WcsNDMap(geom)

    for kn, norm_value in enumerate(norm.center):
        for kp, cste_value in enumerate(cste.center):
            nd_map.data[kp, kn, :, :] = norm_value * map_.data + cste_value

    # template = TemplateNDSpatialModel(nd_map, interp_kwargs={"values_scale": "lin"})
    with caplog.at_level(logging.WARNING):
        template = TemplateNDSpatialModel(nd_map, interp_kwargs={"values_scale": "lin"})
        assert (
            'The filename is not defined. Therefore, the model will not be serialised correctly. To set the filename, the "template_model.filename" attribute can be used.'
            in [_.message for _ in caplog.records]
        )
    assert len(template.parameters) == 2
    assert_allclose(template.parameters["norm"].value, 5)
    assert_allclose(template.parameters["cste"].value, 0)
    assert_allclose(
        template.evaluate(
            geom2d.center_skydir.ra, geom2d.center_skydir.dec, norm=0, cste=0
        ),
        [0],
    )
    assert_allclose(template.evaluate_geom(geom2d), 5 * map_.data, rtol=0.03, atol=10)

    template.parameters["norm"].value = 2
    template.parameters["cste"].value = 0
    assert_allclose(template.evaluate_geom(geom2d), 2 * map_.data, rtol=0.03, atol=10)
    template.filename = str(tmpdir / "template_ND.fits")
    template.write()
    dict_ = template.to_dict()
    template_new = TemplateNDSpatialModel.from_dict(dict_)
    assert_allclose(template_new.map.data, nd_map.data)
    assert len(template_new.parameters) == 2
    assert template_new.parameters["norm"].value == 2
    assert template_new.parameters["cste"].value == 0


@requires_data()
def test_templatespatial_write(tmpdir, caplog):
    filename = "$GAMMAPY_DATA/catalogs/fermi/Extended_archive_v18/Templates/RXJ1713_2016_250GeV.fits"
    map_ = Map.read(filename)
    with caplog.at_level(logging.WARNING):
        template = TemplateSpatialModel(map_)
        assert (
            'The filename is not defined. Therefore, the model will not be serialised correctly. To set the filename, the "template_model.filename" attribute can be used.'
            in [_.message for _ in caplog.records]
        )
    template.filename = filename
    filename_new = str(tmpdir / "template_test.fits")
    template.write(overwrite=True, filename=filename_new)
    assert Path(filename_new).is_file()


@requires_data()
def test_template_spatial_parameters_copy():
    filename = "$GAMMAPY_DATA/catalogs/fermi/Extended_archive_v18/Templates/RXJ1713_2016_250GeV.fits"
    model = TemplateSpatialModel.read(filename, normalize=False)
    model.position = SkyCoord(0, 0, unit="deg", frame="galactic")
    model_copy = model.copy()
    assert_allclose(model.parameters.value, model_copy.parameters.value)



"""
Bayesian analysis with nested sampling
======================================

A demonstration of a Bayesian analysis using the nested sampling technique.

"""


######################################################################
# Context
# -------
#
# 1. Bayesian analysis
# ~~~~~~~~~~~~~~~~~~~~
#
# Bayesian inference uses prior knowledge, in the form of a prior
# distribution, in order to estimate posterior probabilities which we
# traditionally visualise in the form of corner plots. These distributions
# contain more information than a maximum likelihood fit as they reveal not
# only the “best model” but provide a more accurate representation of errors and
# correlation between parameters. In particular, non-Gaussian degeneracies are
# complex to estimate with a maximum likelihood approach.
#
# 2. Limitations of the Markov Chain Monte Carlo approach
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# A well-known approach to estimate this posterior distribution is the
# Markov Chain Monte Carlo (MCMC). This uses an ensemble of walkers to
# produce a chain of samples that after a convergence period will reach a
# stationary state. *Once convergence* is reached, the successive elements
# of the chain are samples of the target posterior distribution. However,
# the weakness of the MCMC approach lies in the "*Once convergence*" part.
# If the walkers are started far from the best likelihood region, the convergence time can be
# long or never reached if the walkers fall in a local minima. The choice
# of the initialisation point can become critical for complex models with
# a high number of dimensions and the ability of these walkers to escape a
# local minimum or to accurately describe a complex likelihood space is
# not guaranteed.
#
# 3. Nested sampling approach
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# To overcome these issues, the nested sampling (NS) algorithm has
# gained traction in physics and astronomy. It is a Monte Carlo
# algorithm for computing an integral of the likelihood function over
# the prior model parameter space introduced in
# `Skilling, 2004 <https://ui.adsabs.harvard.edu/abs/2004AIPC..735..395S>`__.
# The method performs this integral by evolving a collection of points
# through the parameter space (see recent reviews from `Ashton et al.,
# 2022 <https://ui.adsabs.harvard.edu/abs/2022NRvMP...2...39A>`__, and
# `Buchner, 2023 <http://arxiv.org/abs/2101.09675>`__). Without going
# into too many details, one important specificity of the NS method is
# that it starts from the entire parameter space and evolves a
# collection of live points to map all minima (including multiple modes
# if any), whereas Markov Chain Monte Carlo methods require an
# initialisation point and the walkers will explore the local
# likelihood. The ability of these walkers to escape a local minimum or
# to accurately describe a complex likelihood space is not guaranteed.
# This is a fundamental difference with MCMC or Minuit which will
# only ever probe the vicinity along their minimisation paths and do not
# have an overview of the global likelihood landscape. The analysis
# using the NS framework is more CPU time consuming than a standard
# classical fit, but it provides the full posterior distribution for all
# parameters, which is out of reach with traditional fitting techniques
# (N*(N-1)/2 contour plots to generate). In addition, it is more robust
# to the choice of initialisation, requires less human intervention and
# is therefore readily integrated in pipeline analysis. In Gammapy, we
# used the NS implementation of the UltraNest package
# (see `here <https://johannesbuchner.github.io/UltraNest/>`__ for more information), one of the
# leading package in Astronomy (already used in Cosmology and in
# X-rays).
# For a nice visualisation of the NS method see here : `sampling
# visualisation <https://johannesbuchner.github.io/UltraNest/method.html#visualisation>`__.
# And for a tutorial of UltraNest applied to X-ray fitting with concrete examples and questions see : `BXA
# Tutorial <https://peterboorman.com/tutorial_bxa.html>`__.
#
#
# **Note: please cite UltraNest if used for a paper**
#
# If you are using the "UltraNest" library for a paper, please follow its citation scheme:
# `Cite UltraNest <https://johannesbuchner.github.io/UltraNest/issues.html#how-should-i-cite-ultranest>`__.
#
#

######################################################################
# Proposed approach
# -----------------
#
# In this example, we will perform a Bayesian analysis with multiple 1D
# spectra of the Crab nebula data and investigate their posterior
# distributions.
#


######################################################################
# Setup
# -----
#
# As usual, we’ll start with some setup …
#

import matplotlib.pyplot as plt
import numpy as np
import astropy.units as u
from gammapy.datasets import Datasets
from gammapy.datasets import SpectrumDatasetOnOff

from gammapy.modeling.models import (
    SkyModel,
    UniformPrior,
    LogUniformPrior,
)

from gammapy.modeling.sampler import Sampler


######################################################################
# Loading the spectral datasets
# -----------------------------
#
# Here we will load a few Crab 1D spectral data for which we will do a
# fit.
#

path = "$GAMMAPY_DATA/joint-crab/spectra/hess/"

datasets = Datasets()
for id in ["23526", "23559", "23592"]:
    dataset = SpectrumDatasetOnOff.read(f"{path}pha_obs{id}.fits")
    datasets.append(dataset)


######################################################################
# Model definition
# ----------------
#
# Now we want to define the spectral model that will be fitted to the
# data.
# The Crab spectra will be fitted here with a simple powerlaw for
# simplicity.
#

model = SkyModel.create(spectral_model="pl", name="crab")


######################################################################
#
# .. WARNING:: Priors definition:
#    Unlike a traditional fit where priors on the
#    parameters are optional, here it is inherent to the Bayesian approach and
#    are therefore mandatory.
#
# In this case we will set (min,max) prior that will define the
# boundaries in which the sampling will be performed.
# Note that it is usually recommended to use a `~gammapy.modeling.models.LogUniformPrior` for
# the parameters that have a large amplitude range like the
# ``amplitude`` parameter.
# A `~gammapy.modeling.models.UniformPrior` means that the samples will be drawn with uniform
# probability between the (min,max) values in the linear or log space
# in the case of a `~gammapy.modeling.models.LogUniformPrior`.
#

model.spectral_model.amplitude.prior = LogUniformPrior(min=1e-12, max=1e-10)
model.spectral_model.index.prior = UniformPrior(min=1, max=5)
datasets.models = [model]
print(datasets.models)


######################################################################
# Defining the sampler and options
# --------------------------------
#
# As for the `~gammapy.modeling.Fit` object, the `~gammapy.modeling.Sampler` object can receive
# different backend (although just one is available for now).
# The `~gammapy.modeling.Sampler` comes with “reasonable” default parameters, but you can
# change them via the ``sampler_opts`` dictionary.
# Here is a short description of the most relevant parameters that you
# could change :
#
# -  ``live_points``: minimum number of live points throughout the run.
#    More points allow to discover multiple peaks if existing, but is
#    slower. To test the Prior boundaries and for debugging, a lower
#    number (~100) can be used before a production run with more points
#    (~400 or more).
# -  ``frac_remain``: the cut-off condition for the integration, set by the maximum
#    allowed fraction of posterior mass left in the live points vs the dead points. High
#    values (e.g., 0.5) are faster and can be used if the posterior
#    distribution is a relatively simple shape. A low value (1e-1, 1e-2)
#    is optimal for finding peaks, but slower.
# -  ``log_dir``: directory where the output files will be stored.
#    If set to None, no files will be written. If set to a string, a directory
#    will be created containing the ongoing status of the run and final results.
#    For time consuming analysis, it is highly recommended to use that option to monitor
#    the run and restart it in case of a crash (with ``resume=True``).
#
# **Important note:** unlike the MCMC method, you don’t need to define the
# number of steps for which the sampler will run. The algorithm will
# automatically stop once a convergence criteria has been reached.
#

sampler_opts = {
    "live_points": 300,
    "frac_remain": 0.3,
    "log_dir": None,
}

sampler = Sampler(backend="ultranest", sampler_opts=sampler_opts)


######################################################################
# Next we can run the sampler on a given dataset.
# No options are accepted in the run method.
#

result_joint = sampler.run(datasets)


######################################################################
# Understanding the outputs
# -------------------------
#
# In the Jupyter notebook, you should be able to see an interactive
# visualisation of how the parameter space shrinks which starts from the
# (min,max) shrinks down towards the optimal parameters.
#
# The output above is filled with interesting information. Here we
# provide a short description of the most relevant information provided
# above.
# For more detailed information see the `UltraNest
# docs <https://johannesbuchner.github.io/UltraNest/issues.html#what-does-the-status-line-mean>`__.
#
# **During the sampling**
#
# ``Z=-68.8(0.53%) | Like=-63.96..-58.75 [-63.9570..-63.9539]*| it/evals=640/1068 eff=73.7327% N=300``
#
# Some important information here is:
#
# -  Progress (0.53%): the completed fraction of the integral. This is not a time progress bar.
#    Stays at zero for a good fraction of the run.
#
# -  Efficiency (eff value) of the sampling: this indicates out of the proposed new points,
#    how many were accepted. If your efficiency is too small (<<1%), maybe
#    you should revise your priors (e.g use a LogUniform prior for the
#    normalisation).
#
# **Final outputs**
#
# The final lines indicate that all three “convergence” strategies are
# satisfied (samples, posterior uncertainty, and evidence uncertainty).
#
# ``logZ = -65.104 +- 0.292``
#
# The main goal of the Nested sampling algorithm is to estimate Z (the
# Bayesian evidence) which is given above together with an uncertainty.
# In a similar way to deltaLogLike and deltaAIC, deltaLogZ values can be
# used for model comparison.
# For more information see : `on the use of the evidence for model comparison
# <https://ned.ipac.caltech.edu/level5/Sept13/Trotta/Trotta4.html>`__.
# An interesting comparison of the efficiency and false discovery rate of
# model selection with deltaLogLike and deltaLogZ is given in Appendix C of
# `Buchner et al., 2014 <https://ui.adsabs.harvard.edu/abs/2014A%26A...564A.125B>`__.
#
# **Results stored on disk**
#
# if ``log_dir`` is set to a name where the results will be stored, then
# a directory is created containing many useful results and plots.
# A description of these outputs is given in the `Ultranest
# docs <https://johannesbuchner.github.io/UltraNest/performance.html#output-files>`__.
#


######################################################################
# Results
# -------
#


######################################################################
# Within a Bayesian analysis, the concept of best-fit has to be viewed
# differently from what is done in a gradient descent fit.
#
# The output of the Bayesian analysis is the posterior distribution and
# there is no “best-fit” output.
# One has to define, based on the posteriors, what we want to consider
# as “best-fit” and several options are possible:
#
# -  the mean of the distribution
# -  the median
# -  the lowest likelihood value
#
# By default the `~gammapy.modeling.models.DatasetModels` will be updated with the ``mean`` of
# the posterior distributions.
#

print(result_joint.models)


######################################################################
# The `~gammapy.modeling.Sampler` class returns a very rich dictionary.
# The most “standard” information about the posterior distributions can
# be found in :
#

print(result_joint.sampler_results["posterior"])


######################################################################
# Besides mean, errors, etc, an interesting value is the
# ``information gain`` which estimates how much the posterior
# distribution has shrinked with respect to the prior (i.e. how much
# we’ve learned). A value < 1 means that the parameter is poorly
# constrained within the prior range (we haven't learned much with respect to our prior assumption).
# For a physical example see this
# `example <https://arxiv.org/abs/2205.00009>`__.
#
# The `~gammapy.modeling.SamplerResult` dictionary contains also other interesting
# information :
#

print(result_joint.sampler_results.keys())


######################################################################
# Of particular interest, the samples used in the process to approximate
# the posterior distribution can be accessed via :
#

for i, n in enumerate(model.parameters.free_parameters.names):
    s = result_joint.samples[:, i]
    fig, ax = plt.subplots()
    ax.hist(s, bins=30)
    ax.axvline(np.mean(s), ls="--", color="red")
    ax.set_xlabel(n)
    plt.show()


######################################################################
# While the above plots are interesting, the real strength of the Bayesian
# analysis is to visualise all parameters correlations which is usually
# done using “corner plots”.
# Ultranest corner plot function is a wrapper around the `corner
# <https://corner.readthedocs.io/en/latest/api>`__ package.
# See the above link for optional keywords.
# Other packages exist for corner plots, like
# `chainconsumer <https://samreay.github.io/ChainConsumer>`__ which is discussed later in this tutorial.

from ultranest.plot import cornerplot

cornerplot(
    result_joint.sampler_results,
    plot_datapoints=True,
    plot_density=True,
    bins=20,
    title_fmt=".2e",
    smooth=False,
)
plt.show()

# sphinx_gallery_thumbnail_number = 3

######################################################################
# Spectral model error band from samples
# --------------------------------------
#
# To compute the spectral error band ("butterfly plots"), we will directly
# use the samples of the posterior distribution. This is more robust as
# compared to the traditional method of using the covariance matrix of
# the parameters which implicitly assumes Gaussian errors while for the posterior
# distribution there is no shape assumed.
# This difference can become significant when the parameter errors are non-Gaussian.
# For this we will need to convert the list of samples back to the spectral
# model parameters with the relevant units (e.g. normalisation units).


def get_samples_from_posterior(spectral_model, results):
    """
    Create a list of spectral parameters with correct units
    from the unitless parameters returned by the sampler.
    """
    n_samples = results.samples.shape[0]
    samples = []
    for p in spectral_model.parameters:
        try:
            idx = spectral_model.parameters.free_unique_parameters.index(p)
            samples.append(results.samples[:, idx] * p.unit)
        except ValueError:
            samples.append(np.ones(n_samples) * p.quantity)
    return samples


samples = get_samples_from_posterior(datasets.models[0].spectral_model, result_joint)

######################################################################
# Next we can provide these samples to the `~gammapy.modeling.models.SpectralModel.plot_error`
# method.

datasets.models[0].spectral_model.plot_error(
    energy_bounds=[0.5 * u.TeV, 50 * u.TeV], sed_type="e2dnde", samples=samples
)
plt.show()

######################################################################
# Individual run analysis
# -----------------------
#
# Now we’ll analyse several Crab runs individually so that we can compare
# them.
#

result_0 = sampler.run(datasets[0])
result_1 = sampler.run(datasets[1])
result_2 = sampler.run(datasets[2])


######################################################################
# Comparing the posterior distribution of all runs
# ------------------------------------------------
#
# For a comparison of different posterior distributions, we can use the
# package chainconsumer.
# As this is not a Gammapy dependency, you’ll need to install it.
# More info here : https://samreay.github.io/ChainConsumer/
#

# Uncomment this if you have installed `chainconsumer`.

# from chainconsumer import Chain, ChainConfig, ChainConsumer, PlotConfig, Truth, make_sample
# from pandas import DataFrame
# c = ChainConsumer()
# def create_chain(result, name, color="k"):
#    return Chain(
#        samples=DataFrame(result, columns=["index", "amplitude"]),
#        name=name,
#        color=color,
#        smooth=7,
#        shade=False,
#        linewidth=1.0,
#        cmap="magma",
#        show_contour_labels=True,
#        kde= True
#    )
# c.add_chain(create_chain(result_joint.samples, "joint"))
# c.add_chain(create_chain(result_0.samples, "run0", "g"))
# c.add_chain(create_chain(result_1.samples, "run1", "b"))
# c.add_chain(create_chain(result_2.samples, "run2", "y"))
# fig = c.plotter.plot()
# plt.show()


######################################################################
#
# Corner plot comparison
# ----------------------
#
# .. figure:: ../../_static/cornerplot-multiple-runs-Crab.png
#     :alt: Corner plot of Crab runs
#
#     Corner plot comparing the three Crab runs.
#
#
# We can see the joint analysis allows to better constrain the
# parameters than the individual runs (more observation time is of
# course better).
# One can note as well that one of the run has a notably different
# amplitude (possibly due to calibrations or/and atmospheric issues).
#



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from astropy.coordinates import SkyCoord
from regions import PolygonSkyRegion
import matplotlib as mpl
import matplotlib.pyplot as plt


def containment_region(map_, fraction=0.393, apply_union=True):
    """Find the iso-contours region corresponding to a given containment for a map of integral quantities with a flat geometry.

    Parameters
    ----------
    map_ : `~gammapy.maps.WcsNDMap`
        Map of integral quantities.
    fraction : float, optional
        Containment fraction. Default is 0.393.
    apply_union : bool, optional
        It True return a compound region otherwise return a list of polygon regions.
        Default is True. Note that compound regions cannot be written in ds9 format but can always be saved with numpy.savez.

    Returns
    -------
    regions : list of `~regions.PolygonSkyRegion` or `~regions.CompoundSkyRegion`
        Regions from iso-contours matching containment fraction.
    """
    from . import WcsNDMap

    if not isinstance(map_, WcsNDMap):
        raise TypeError(
            f"This function is only supported for WcsNDMap, got {type(map_)} instead."
        )

    fmax = np.nanmax(map_.data)
    if fmax > 0.0:
        ordered = np.sort(map_.data.flatten())[::-1]
        cumsum = np.nancumsum(ordered)
        ind = np.argmin(np.abs(cumsum / cumsum.max() - fraction))
        fval = ordered[ind]

        plt.ioff()
        fig = plt.figure()
        cs = plt.contour(map_.data.squeeze(), [fval])
        plt.close(fig)
        plt.ion()
        regions_pieces = []

        try:
            paths_all = cs.get_paths()[0]
            starts = np.where(paths_all.codes == 1)[0]
            stops = np.where(paths_all.codes == 79)[0] + 1
            paths = []
            for start, stop in zip(starts, stops):
                paths.append(
                    mpl.path.Path(
                        paths_all.vertices[start:stop],
                        codes=paths_all.codes[start:stop],
                    )
                )
        except AttributeError:
            paths = cs.collections[0].get_paths()
        for pp in paths:
            vertices = []
            for v in pp.vertices:
                v_coord = map_.geom.pix_to_coord(v)
                vertices.append([v_coord[0], v_coord[1]])
            vertices = SkyCoord(vertices, frame=map_.geom.frame)
            regions_pieces.append(PolygonSkyRegion(vertices))

        if apply_union:
            regions_union = regions_pieces[0]
            for region in regions_pieces[1:]:
                regions_union = regions_union.union(region)
            return regions_union
        else:
            return regions_pieces
    else:
        raise ValueError("No positive values in the map.")


def containment_radius(map_, fraction=0.393, position=None):
    """Compute containment radius from a position in a map with integral quantities and flat geometry.

    Parameters
    ----------
    map_ : `~gammapy.maps.WcsNDMap`
        Map of integral quantities.
    fraction : float
        Containment fraction. Default is 0.393.
    position : `~astropy.coordinates.SkyCoord`
        Position from where the containment is computed.
        Default is the center of the Map.

    Returns
    -------
    radius : `~astropy.coordinates.Angle`
        Minimal radius required to reach the given containment fraction.

    """
    from . import WcsNDMap

    if not isinstance(map_, WcsNDMap):
        raise TypeError(
            f"This function is only supported for WcsNDMap, got {type(map_)} instead."
        )

    if position is None:
        position = map_.geom.center_skydir

    fmax = np.nanmax(map_.data)
    if fmax > 0.0:
        sep = map_.geom.separation(position).flatten()
        order = np.argsort(sep)
        ordered = map_.data.flatten()[order]
        cumsum = np.nancumsum(ordered)
        ind = np.argmin(np.abs(cumsum / cumsum.max() - fraction))
    else:
        raise ValueError("No positive values in the map.")
    return sep[order][ind]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from gammapy.astro.darkmatter import profiles
from gammapy.utils.testing import assert_quantity_allclose

dm_profiles = [
    profiles.ZhaoProfile,
    profiles.NFWProfile,
    profiles.EinastoProfile,
    profiles.IsothermalProfile,
    profiles.BurkertProfile,
    profiles.MooreProfile,
]


@pytest.mark.parametrize("profile", dm_profiles)
def test_profiles(profile):
    p = profile()
    p.scale_to_local_density()
    actual = p(p.DISTANCE_GC)
    desired = p.LOCAL_DENSITY

    assert_quantity_allclose(actual, desired)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import numpy as np
from astropy.table import Table
from astropy.utils import lazyproperty
from gammapy.utils.fits import HDULocation
from gammapy.utils.scripts import make_path

__all__ = ["HDUIndexTable"]

log = logging.getLogger(__name__)


class HDUIndexTable(Table):
    """HDU index table.

    See `HDU index table <https://gamma-astro-data-formats.readthedocs.io/en/latest/data_storage/hdu_index/index.html#hdu-index>`__.
    """

    VALID_HDU_TYPE = [
        "events",
        "gti",
        "aeff",
        "edisp",
        "psf",
        "bkg",
        "rad_max",
        "pointing",
    ]
    """Valid values for `HDU_TYPE`."""

    VALID_HDU_CLASS = [
        "events",
        "gti",
        "aeff_2d",
        "edisp_2d",
        "psf_table",
        "psf_3gauss",
        "psf_king",
        "bkg_2d",
        "bkg_3d",
        "rad_max_2d",
        "pointing",
    ]
    """Valid values for `HDU_CLASS`."""

    @classmethod
    def read(cls, filename, **kwargs):
        """Read `HDU index table <https://gamma-astro-data-formats.readthedocs.io/en/latest/data_storage/hdu_index/index.html#hdu-index>`__.

        Parameters
        ----------
        filename : `pathlib.Path` or str
            Filename.
        **kwargs : dict, optional
            Keyword arguments passed to `~astropy.table.Table.read`.
        """
        filename = make_path(filename)
        table = super().read(filename, **kwargs)
        table.meta["BASE_DIR"] = filename.parent.as_posix()

        # TODO: this is a workaround for the joint-crab validation with astropy>4.0.
        # TODO: Remove when handling of empty columns is clarified
        table["FILE_DIR"].fill_value = ""

        return table.filled()

    @property
    def base_dir(self):
        """Base directory."""
        return make_path(self.meta.get("BASE_DIR", ""))

    def hdu_location(self, obs_id, hdu_type=None, hdu_class=None, warn_missing=True):
        """Create `HDULocation` for a given selection.

        Parameters
        ----------
        obs_id : int
            Observation ID.
        hdu_type : str, optional
            HDU type (see `~gammapy.data.HDUIndexTable.VALID_HDU_TYPE`). Default is None.
        hdu_class : str, optional
            HDU class (see `~gammapy.data.HDUIndexTable.VALID_HDU_CLASS`). Default is None.
        warn_missing : bool, optional
            Warn if no HDU is found matching the selection. Default is True.

        Returns
        -------
        location : `~gammapy.data.HDULocation`
            HDU location.
        """
        self._validate_selection(obs_id=obs_id, hdu_type=hdu_type, hdu_class=hdu_class)

        idx = self.row_idx(obs_id=obs_id, hdu_type=hdu_type, hdu_class=hdu_class)

        if len(idx) == 1:
            idx = idx[0]
        elif len(idx) == 0:
            if warn_missing:
                log.warning(
                    f"No HDU found matching: OBS_ID = {obs_id}, HDU_TYPE = {hdu_type},"
                    " HDU_CLASS = {hdu_class}"
                )
            return None
        else:
            idx = idx[0]
            log.warning(
                f"Found multiple HDU matching: OBS_ID = {obs_id}, HDU_TYPE = {hdu_type},"
                " HDU_CLASS = {hdu_class}."
                f" Returning the first entry, which has "
                f"HDU_TYPE = {self[idx]['HDU_TYPE']} and HDU_CLASS = {self[idx]['HDU_CLASS']}"
            )

        return self.location_info(idx)

    def _validate_selection(self, obs_id, hdu_type, hdu_class):
        """Validate HDU selection.

        The goal is to give helpful error messages to the user.
        """
        if hdu_type is None and hdu_class is None:
            raise ValueError("You have to specify `hdu_type` or `hdu_class`.")

        if hdu_type and hdu_type not in self.VALID_HDU_TYPE:
            valid = [str(_) for _ in self.VALID_HDU_TYPE]
            raise ValueError(f"Invalid hdu_type: {hdu_type}. Valid values are: {valid}")

        if hdu_class and hdu_class not in self.VALID_HDU_CLASS:
            valid = [str(_) for _ in self.VALID_HDU_CLASS]
            raise ValueError(
                f"Invalid hdu_class: {hdu_class}. Valid values are: {valid}"
            )

        if obs_id not in self["OBS_ID"]:
            raise IndexError(f"No entry available with OBS_ID = {obs_id}")

    def row_idx(self, obs_id, hdu_type=None, hdu_class=None):
        """Table row indices for a given selection.

        Parameters
        ----------
        obs_id : int
            Observation ID.
        hdu_type : str, optional
            HDU type (see `~gammapy.data.HDUIndexTable.VALID_HDU_TYPE`). Default is None.
        hdu_class : str, optional
            HDU class (see `~gammapy.data.HDUIndexTable.VALID_HDU_CLASS`). Default is None.

        Returns
        -------
        idx : list of int
            List of row indices matching the selection.
        """
        selection = self["OBS_ID"] == obs_id

        if hdu_class:
            is_hdu_class = self._hdu_class_stripped == hdu_class
            selection &= is_hdu_class

        if hdu_type:
            is_hdu_type = self._hdu_type_stripped == hdu_type
            selection &= is_hdu_type

        idx = np.where(selection)[0]
        return list(idx)

    def location_info(self, idx):
        """Create `HDULocation` for a given row index."""
        row = self[idx]
        return HDULocation(
            hdu_class=row["HDU_CLASS"].strip(),
            base_dir=self.base_dir.as_posix(),
            file_dir=row["FILE_DIR"].strip(),
            file_name=row["FILE_NAME"].strip(),
            hdu_name=row["HDU_NAME"].strip(),
        )

    @lazyproperty
    def _hdu_class_stripped(self):
        return np.array([_.strip() for _ in self["HDU_CLASS"]])

    @lazyproperty
    def _hdu_type_stripped(self):
        return np.array([_.strip() for _ in self["HDU_TYPE"]])

    @lazyproperty
    def obs_id_unique(self):
        """Observation IDs (unique)."""
        return np.unique(np.sort(self["OBS_ID"]))

    @lazyproperty
    def hdu_type_unique(self):
        """HDU types (unique)."""
        return list(np.unique(np.sort([_.strip() for _ in self["HDU_TYPE"]])))

    @lazyproperty
    def hdu_class_unique(self):
        """HDU classes (unique)."""
        return list(np.unique(np.sort([_.strip() for _ in self["HDU_CLASS"]])))

    def summary(self):
        """Summary report as a string."""
        obs_id = self.obs_id_unique
        return (
            "HDU index table:\n"
            f"BASE_DIR: {self.base_dir}\n"
            f"Rows: {len(self)}\n"
            f"OBS_ID: {obs_id[0]} -- {obs_id[-1]}\n"
            f"HDU_TYPE: {self.hdu_type_unique}\n"
            f"HDU_CLASS: {self.hdu_class_unique}\n"
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
from scipy.stats import ncx2


__all__ = ["sigma_to_ts", "ts_to_sigma"]


def sigma_to_ts(n_sigma, df=1, n_sigma_asimov=0):
    """Convert number of sigma to delta ts.

    Assumes that the TS follows a chi2 distribution according to Wilks theorem [1].
    This is valid only if:

    - the two hypotheses tested can be defined in the same parameters space
    - the true value is not at the boundary of this parameters space.

    Parameters
    ----------
    n_sigma : float
        Significance in number of sigma.
    df : int, optional
        Number of degree of freedom. Default is 1.
    n_sigma_asimov : float, optional
        Significance in number of sigma in the Asimov dataset
        (in which counts are equal to the predicted counts).
        In that case the function applies the Wald test described in [2] and [3],
        where the TS of H1 under the H0 assumption is assumed to follow a non-central chi2 distribution.
        Should only be used for sensitivity computations. Default is 0.

    Returns
    -------
    ts : float
        Test statistic value.

    References
    ----------
    .. [1] Wilks theorem: https://en.wikipedia.org/wiki/Wilks%27_theorem

    .. [2] Wald (1943): https://www.pp.rhul.ac.uk/~cowan/stat/wald1943.pdf

    .. [3] Cowan et al. (2011), European Physical Journal C, 71, 1554.
        doi:10.1140/epjc/s10052-011-1554-0.
    """
    ts_asimov = n_sigma_asimov**2
    p_value = ncx2.sf(n_sigma**2, df=1, nc=ts_asimov)
    return ncx2.isf(p_value, df=df, nc=ts_asimov)


def ts_to_sigma(ts, df=1, ts_asimov=0):
    """Convert delta ts to number of sigma.

    Assumes that the TS follows a chi2 distribution according to Wilks theorem [1].
    This is valid only if:

    - the two hypotheses tested can be defined in the same parameters space
    - the true value is not at the boundary of this parameters space.

    Parameters
    ----------
    ts : float
        Test statistic value.
    df : int, optional
        Number of degree of freedom. Default is 1.
    ts_asimov : float, optional
        TS value in the Asimov dataset
        (in which counts equal to the predicted counts).
        In that case the function applies the Wald test described in [2] and [3],
        and the TS is assumed to follow a non-central chi2 distribution.
        Should only be used for sensitivity computations. Default is 0.


    Returns
    -------
    n_sigma : float
        Significance in number of sigma.

    References
    ----------
    .. [1] Wilks theorem: https://en.wikipedia.org/wiki/Wilks%27_theorem

    .. [2] Wald (1943): https://www.pp.rhul.ac.uk/~cowan/stat/wald1943.pdf

    .. [3] Cowan et al. (2011), European Physical Journal C, 71, 1554.
        doi:10.1140/epjc/s10052-011-1554-0.
    """
    p_value = ncx2.sf(ts, df=df, nc=ts_asimov)
    return np.sqrt(ncx2.isf(p_value, df=1, nc=ts_asimov))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Implementation of adaptive smoothing algorithms."""

import numpy as np
from astropy.convolution import Gaussian2DKernel, Tophat2DKernel
from astropy.coordinates import Angle
from gammapy.datasets import MapDatasetOnOff
from gammapy.maps import Map, Maps, WcsNDMap
from gammapy.modeling.models import PowerLawSpectralModel
from gammapy.stats import CashCountsStatistic
from gammapy.utils.array import scale_cube
from gammapy.utils.pbar import progress_bar
from ..core import Estimator
from ..utils import estimate_exposure_reco_energy

__all__ = ["ASmoothMapEstimator"]


def _sqrt_ts_asmooth(counts, background):
    """Significance according to formula (5) in asmooth paper."""
    return (counts - background) / np.sqrt(counts + background)


class ASmoothMapEstimator(Estimator):
    """Adaptively smooth counts image.

    Achieves a roughly constant sqrt(TS) of features across the whole image.

    Algorithm based on [1]_.

    The algorithm was slightly adapted to also allow Li & Ma  to estimate the
    sqrt(TS) of a feature in the image.

    Parameters
    ----------
    scales : `~astropy.units.Quantity`
        Smoothing scales.
    kernel : `astropy.convolution.Kernel`
        Smoothing kernel.
    spectral_model : `~gammapy.modeling.models.SpectralModel`, optional
        Spectral model assumption. Default is power-law with spectral index of 2.
    method : {'lima', 'asmooth'}
        Significance estimation method. Default is 'lima'.
    threshold : float
        Significance threshold. Default is 5.
    energy_edges : list of `~astropy.units.Quantity`, optional
        Edges of the target maps energy bins. The resulting bin edges won't be exactly equal to the input ones,
        but rather the closest values to the energy axis edges of the parent dataset.
        Default is None: apply the estimator in each energy bin of the parent dataset.
        For further explanation see :ref:`estimators`.

    Examples
    --------
    >>> import astropy.units as u
    >>> import numpy as np
    >>> from gammapy.estimators import ASmoothMapEstimator
    >>> from gammapy.datasets import MapDataset
    >>> dataset = MapDataset.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
    >>> scales = u.Quantity(np.arange(0.1, 1, 0.1), unit="deg")
    >>> smooth = ASmoothMapEstimator(threshold=3, scales=scales, energy_edges=[1, 10] * u.TeV)
    >>> images = smooth.run(dataset)

    References
    ----------
    .. [1] `Ebeling et al. (2006), "ASMOOTH: a simple and efficient algorithm for adaptive kernel smoothing of
      two-dimensional imaging data" <https://ui.adsabs.harvard.edu/abs/2006MNRAS.368…65E>`_
    """

    tag = "ASmoothMapEstimator"

    def __init__(
        self,
        scales=None,
        kernel=Gaussian2DKernel,
        spectral_model=None,
        method="lima",
        threshold=5,
        energy_edges=None,
    ):
        if spectral_model is None:
            spectral_model = PowerLawSpectralModel(index=2)

        self.spectral_model = spectral_model

        if scales is None:
            scales = self.get_scales(n_scales=9, kernel=kernel)

        self.scales = scales
        self.kernel = kernel
        self.threshold = threshold
        self.method = method
        self.energy_edges = energy_edges

    def selection_all(self):
        """Which quantities are computed."""
        return

    @staticmethod
    def get_scales(n_scales, factor=np.sqrt(2), kernel=Gaussian2DKernel):
        """Create list of Gaussian widths.

        Parameters
        ----------
        n_scales : int
            Number of scales.
        factor : float
            Incremental factor.

        Returns
        -------
        scales : `~numpy.ndarray`
            Scale array.
        """
        if kernel == Gaussian2DKernel:
            sigma_0 = 1.0 / np.sqrt(9 * np.pi)
        elif kernel == Tophat2DKernel:
            sigma_0 = 1.0 / np.sqrt(np.pi)

        return sigma_0 * factor ** np.arange(n_scales)

    def get_kernels(self, pixel_scale):
        """Get kernels according to the specified method.

        Parameters
        ----------
        pixel_scale : `~astropy.coordinates.Angle`
            Sky image pixel scale.

        Returns
        -------
        kernels : list
            List of `~astropy.convolution.Kernel`.
        """
        scales = self.scales.to_value("deg") / Angle(pixel_scale).deg

        kernels = []
        for scale in scales:  # .value:
            kernel = self.kernel(scale, mode="oversample")
            # TODO: check if normalizing here makes sense
            kernel.normalize("peak")
            kernels.append(kernel)

        return kernels

    @staticmethod
    def _sqrt_ts_cube(cubes, method):
        if method in {"lima"}:
            scube = CashCountsStatistic(cubes["counts"], cubes["background"]).sqrt_ts
        elif method == "asmooth":
            scube = _sqrt_ts_asmooth(cubes["counts"], cubes["background"])
        elif method == "ts":
            raise NotImplementedError()
        else:
            raise ValueError(
                "Not a valid sqrt_ts estimation method."
                " Choose one of the following: 'lima' or 'asmooth'"
            )
        return scube

    def run(self, dataset):
        """Run adaptive smoothing on input MapDataset.

        Notes
        -----
        The progress bar can be displayed for this function.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset` or `~gammapy.datasets.MapDatasetOnOff`
            The input dataset (with one bin in energy at most).

        Returns
        -------
        images : dict of `~gammapy.maps.WcsNDMap`
            Smoothed images; keys are:
                * 'counts'
                * 'background'
                * 'flux' (optional)
                * 'scales'
                * 'sqrt_ts'.
        """
        energy_axis = self._get_energy_axis(dataset)

        results = []

        for energy_min, energy_max in progress_bar(
            energy_axis.iter_by_edges, desc="Energy bins"
        ):
            dataset_sliced = dataset.slice_by_energy(
                energy_min=energy_min, energy_max=energy_max, name=dataset.name
            )
            if dataset.models is not None:
                models_sliced = dataset.models._slice_by_energy(
                    energy_min=energy_min,
                    energy_max=energy_max,
                )
                dataset_sliced.models = models_sliced
            result = self.estimate_maps(dataset_sliced)
            results.append(result)

        maps = Maps()

        for name in results[0].keys():
            maps[name] = Map.from_stack(
                maps=[_[name] for _ in results], axis_name="energy"
            )

        return maps

    def estimate_maps(self, dataset):
        """Run adaptive smoothing on input Maps.

        Parameters
        ----------
        dataset : `MapDataset`
            Dataset.

        Returns
        -------
        images : dict of `~gammapy.maps.WcsNDMap`
            Smoothed images; keys are:
                * 'counts'
                * 'background'
                * 'flux' (optional)
                * 'scales'
                * 'sqrt_ts'.
        """
        dataset_image = dataset.to_image(name=dataset.name)
        dataset_image.models = dataset.models

        # extract 2d arrays
        counts = dataset_image.counts.data[0].astype(float)
        background = dataset_image.npred_background().data[0]

        if isinstance(dataset_image, MapDatasetOnOff):
            background = dataset_image.background.data[0]

        if dataset_image.exposure is not None:
            exposure = estimate_exposure_reco_energy(dataset_image, self.spectral_model)
        else:
            exposure = None

        pixel_scale = dataset_image.counts.geom.pixel_scales.mean()
        kernels = self.get_kernels(pixel_scale)

        cubes = {}
        cubes["counts"] = scale_cube(counts, kernels)
        cubes["background"] = scale_cube(background, kernels)

        if exposure is not None:
            flux = (dataset_image.counts - background) / exposure
            cubes["flux"] = scale_cube(flux.data[0], kernels)

        cubes["sqrt_ts"] = self._sqrt_ts_cube(cubes, method=self.method)

        smoothed = self._reduce_cubes(cubes, kernels)

        result = {}

        geom = dataset_image.counts.geom

        for name, data in smoothed.items():
            # set remaining pixels with sqrt_ts < threshold to mean value
            if name in ["counts", "background"]:
                mask = np.isnan(data)
                data[mask] = np.mean(locals()[name][mask])
                result[name] = WcsNDMap(geom, data, unit="")
            else:
                unit = "deg" if name == "scale" else ""
                result[name] = WcsNDMap(geom, data, unit=unit)

        if exposure is not None:
            data = smoothed["flux"]
            mask = np.isnan(data)
            data[mask] = np.mean(flux.data[0][mask])
            result["flux"] = WcsNDMap(geom, data, unit=flux.unit)

        return result

    def _reduce_cubes(self, cubes, kernels):
        """
        Combine scale cube to image.

        Parameters
        ----------
        cubes : dict
            Data cubes.
        """
        shape = cubes["counts"].shape[:2]
        smoothed = {}

        # Init smoothed data arrays
        for key in ["counts", "background", "scale", "sqrt_ts"]:
            smoothed[key] = np.tile(np.nan, shape)

        if "flux" in cubes:
            smoothed["flux"] = np.tile(np.nan, shape)

        for idx, scale in enumerate(self.scales):
            # slice out 2D image at index idx out of cube
            slice_ = np.s_[:, :, idx]

            mask = np.isnan(smoothed["counts"])
            mask = (cubes["sqrt_ts"][slice_] > self.threshold) & mask

            smoothed["scale"][mask] = scale
            smoothed["sqrt_ts"][mask] = cubes["sqrt_ts"][slice_][mask]

            # renormalize smoothed data arrays
            norm = kernels[idx].array.sum()
            for key in ["counts", "background"]:
                smoothed[key][mask] = cubes[key][slice_][mask] / norm

            if "flux" in cubes:
                smoothed["flux"][mask] = cubes["flux"][slice_][mask] / norm

        return smoothed
